import logging
import typing
from datetime import datetime
from decimal import Decimal
from typing import List, Literal, Optional, Union

import numpy
import xarray
from pandas import Timestamp

from copernicusmarine.catalogue_parser.catalogue_parser import (
    CopernicusMarineDatasetServiceType,
)
from copernicusmarine.catalogue_parser.request_structure import (
    DatasetTimeAndGeographicalSubset,
)
from copernicusmarine.core_functions import custom_open_zarr
from copernicusmarine.core_functions.exceptions import (
    CoordinatesOutOfDatasetBounds,
    MinimumLongitudeGreaterThanMaximumLongitude,
    VariableDoesNotExistInTheDataset,
)
from copernicusmarine.core_functions.models import SubsetMethod
from copernicusmarine.core_functions.utils import (
    ServiceNotSupported,
    convert_datetime64_to_netcdf_timestamp,
)
from copernicusmarine.download_functions.subset_parameters import (
    DepthParameters,
    GeographicalParameters,
    LatitudeParameters,
    LongitudeParameters,
    TemporalParameters,
)

logger = logging.getLogger("copernicus_marine_root_logger")

COORDINATES_LABEL = {
    "latitude": ["latitude", "nav_lat", "x", "lat"],
    "longitude": ["longitude", "nav_lon", "y", "lon"],
    "time": ["time_counter", "time"],
    "depth": ["depth", "deptht", "elevation"],
}


def _dataset_custom_sel(
    dataset: xarray.Dataset,
    coord_type: Literal["latitude", "longitude", "depth", "time"],
    coord_selection: Union[float, slice, datetime, None],
    method: Union[str, None] = None,
) -> xarray.Dataset:
    for coord_label in COORDINATES_LABEL[coord_type]:
        if coord_label in dataset.sizes:
            tmp_dataset = dataset.sel(
                {coord_label: coord_selection}, method=method
            )
            if tmp_dataset.coords[coord_label].size == 0 or (
                coord_label not in tmp_dataset.sizes
            ):
                target = (
                    coord_selection.start
                    if isinstance(coord_selection, slice)
                    else coord_selection
                )
                nearest_neighbor_value = dataset.sel(
                    {coord_label: target}, method="nearest"
                )[coord_label].values
                dataset = dataset.sel(
                    {
                        coord_label: slice(
                            nearest_neighbor_value, nearest_neighbor_value
                        )
                    }
                )
            else:
                dataset = tmp_dataset
    return dataset


def get_size_of_coordinate_subset(
    dataset: xarray.Dataset,
    coordinate: str,
    minimum: Optional[Union[float, datetime]],
    maximum: Optional[Union[float, datetime]],
) -> int:
    for label in COORDINATES_LABEL[coordinate]:
        if label in dataset.sizes:
            return (
                dataset.coords[coordinate]
                .sel({coordinate: slice(minimum, maximum)}, method=None)
                .coords[coordinate]
                .size
            )
    else:
        raise KeyError(
            f"Could not subset on {coordinate}. "
            "Didn't find an equivalent in the dataset."
        )


def _shift_longitude_dimension(
    dataset: xarray.Dataset,
    minimum_longitude_modulus: float,
):
    window = (
        minimum_longitude_modulus + 180
    )  # compute the degrees needed to move the dataset
    for coord_label in COORDINATES_LABEL["longitude"]:
        if coord_label in dataset.sizes:
            attrs = dataset[coord_label].attrs
            if "valid_min" in attrs:
                attrs["valid_min"] += window
            if "valid_max" in attrs:
                attrs["valid_max"] += window
            dataset = dataset.assign_coords(
                {
                    coord_label: (
                        (dataset[coord_label] + (180 - window)) % 360
                    )
                    - (180 - window)
                }
            ).sortby(coord_label)
            dataset[coord_label].attrs = attrs
    return dataset


def _latitude_subset(
    dataset: xarray.Dataset,
    latitude_parameters: LatitudeParameters,
) -> xarray.Dataset:
    minimum_latitude = latitude_parameters.minimum_latitude
    maximum_latitude = latitude_parameters.maximum_latitude
    if minimum_latitude is not None or maximum_latitude is not None:
        latitude_selection = (
            minimum_latitude
            if minimum_latitude == maximum_latitude
            else slice(minimum_latitude, maximum_latitude)
        )
        latitude_method = (
            "nearest" if minimum_latitude == maximum_latitude else None
        )
        dataset = _dataset_custom_sel(
            dataset, "latitude", latitude_selection, latitude_method
        )

    return dataset


def _longitude_subset(
    dataset: xarray.Dataset,
    longitude_parameters: LongitudeParameters,
) -> xarray.Dataset:
    minimum_longitude = longitude_parameters.minimum_longitude
    maximum_longitude = longitude_parameters.maximum_longitude
    longitude_method = None
    if minimum_longitude is not None or maximum_longitude is not None:
        if minimum_longitude is not None and maximum_longitude is not None:
            if minimum_longitude > maximum_longitude:
                raise MinimumLongitudeGreaterThanMaximumLongitude(
                    "--minimum-longitude option must be smaller "
                    "or equal to --maximum-longitude"
                )
            if maximum_longitude - minimum_longitude >= 360:
                longitude_selection: Union[float, slice, None] = None
            elif minimum_longitude == maximum_longitude:
                longitude_selection = longitude_modulus(minimum_longitude)
                longitude_method = "nearest"
            else:
                minimum_longitude_modulus = longitude_modulus(
                    minimum_longitude
                )
                maximum_longitude_modulus = longitude_modulus(
                    maximum_longitude
                )
                if maximum_longitude_modulus < minimum_longitude_modulus:
                    maximum_longitude_modulus += 360
                    dataset = _shift_longitude_dimension(
                        dataset, minimum_longitude_modulus
                    )
                longitude_selection = slice(
                    minimum_longitude_modulus,
                    maximum_longitude_modulus,
                )
        else:
            longitude_selection = slice(minimum_longitude, maximum_longitude)

        if longitude_selection is not None:
            dataset = _dataset_custom_sel(
                dataset, "longitude", longitude_selection, longitude_method
            )
    return dataset


def _temporal_subset(
    dataset: xarray.Dataset,
    temporal_parameters: TemporalParameters,
) -> xarray.Dataset:
    start_datetime = temporal_parameters.start_datetime
    end_datetime = temporal_parameters.end_datetime
    if start_datetime is not None or end_datetime is not None:
        temporal_selection = (
            start_datetime
            if start_datetime == end_datetime
            else slice(start_datetime, end_datetime)
        )
        temporal_method = "nearest" if start_datetime == end_datetime else None
        dataset = _dataset_custom_sel(
            dataset, "time", temporal_selection, temporal_method
        )
    return dataset


def _depth_subset(
    dataset: xarray.Dataset,
    depth_parameters: DepthParameters,
) -> xarray.Dataset:
    def convert_elevation_to_depth(dataset: xarray.Dataset):
        if "elevation" in dataset.sizes:
            attrs = dataset["elevation"].attrs
            dataset = dataset.reindex(elevation=dataset.elevation[::-1])
            dataset["elevation"] = dataset.elevation * (-1)
            attrs["positive"] = "down"
            attrs["standard_name"] = "depth"
            attrs["long_name"] = "Depth"
            attrs["units"] = "m"
            dataset = dataset.rename({"elevation": "depth"})
            dataset.depth.attrs = attrs
        return dataset

    def update_elevation_attributes(dataset: xarray.Dataset):
        if "elevation" in dataset.sizes:
            attrs = dataset["elevation"].attrs
            attrs["positive"] = "up"
            attrs["standard_name"] = "elevation"
            attrs["long_name"] = "Elevation"
            attrs["units"] = "m"
            dataset["elevation"].attrs = attrs
        return dataset

    if depth_parameters.vertical_dimension_as_originally_produced:
        dataset = convert_elevation_to_depth(dataset)
    else:
        dataset = update_elevation_attributes(dataset)
    minimum_depth = depth_parameters.minimum_depth
    maximum_depth = depth_parameters.maximum_depth
    if minimum_depth is not None or maximum_depth is not None:
        coords = (
            dataset.coords if isinstance(dataset, xarray.Dataset) else dataset
        )
        if "elevation" in coords:
            minimum_depth = (
                minimum_depth * -1.0 if minimum_depth is not None else None
            )
            maximum_depth = (
                maximum_depth * -1.0 if maximum_depth is not None else None
            )
            minimum_depth, maximum_depth = maximum_depth, minimum_depth

        depth_selection = (
            minimum_depth
            if minimum_depth == maximum_depth
            else slice(minimum_depth, maximum_depth)
        )
        depth_method = "nearest" if minimum_depth == maximum_depth else None
        dataset = _dataset_custom_sel(
            dataset, "depth", depth_selection, depth_method
        )
    return dataset


def _get_variable_name_from_standard_name(
    dataset: xarray.Dataset, standard_name: str
) -> Optional[str]:
    for variable_name in dataset.variables:
        if (
            hasattr(dataset[variable_name], "standard_name")
            and dataset[variable_name].standard_name == standard_name
        ):
            return str(variable_name)
    return None


def _variables_subset(
    dataset: xarray.Dataset, variables: List[str]
) -> xarray.Dataset:
    dataset_variables_filter = []

    for variable in variables:
        if variable in dataset.variables:
            dataset_variables_filter.append(variable)
        else:
            variable_name_from_standard_name = (
                _get_variable_name_from_standard_name(dataset, variable)
            )
            if variable_name_from_standard_name is not None:
                dataset_variables_filter.append(
                    variable_name_from_standard_name
                )
            else:
                raise VariableDoesNotExistInTheDataset(variable)
    return dataset[numpy.array(dataset_variables_filter)]


def _update_dataset_coordinate_valid_minmax_attributes(
    dataset: xarray.Dataset,
) -> xarray.Dataset:
    for coordinate_label in COORDINATES_LABEL:
        for coordinate_alias in COORDINATES_LABEL[coordinate_label]:
            if coordinate_alias in dataset.sizes:
                coord = dataset[coordinate_alias]
                attrs = coord.attrs
                if "time" in coordinate_label:
                    min_time_dimension = coord.values.min()
                    max_time_dimension = coord.values.max()
                    netcdf_unit = coord.encoding["units"]
                    valid_min = convert_datetime64_to_netcdf_timestamp(
                        min_time_dimension, netcdf_unit
                    )
                    valid_max = convert_datetime64_to_netcdf_timestamp(
                        max_time_dimension, netcdf_unit
                    )
                    attrs["valid_min"] = valid_min
                    attrs["valid_max"] = valid_max
                else:
                    attrs["valid_min"] = coord.values.min()
                    attrs["valid_max"] = coord.values.max()

                coord.attrs = attrs

    return dataset


def subset(
    dataset: xarray.Dataset,
    variables: Optional[List[str]],
    geographical_parameters: GeographicalParameters,
    temporal_parameters: TemporalParameters,
    depth_parameters: DepthParameters,
) -> xarray.Dataset:
    if variables:
        dataset = _variables_subset(dataset, variables)

    dataset = _latitude_subset(
        dataset, geographical_parameters.latitude_parameters
    )
    dataset = _longitude_subset(
        dataset, geographical_parameters.longitude_parameters
    )

    dataset = _temporal_subset(dataset, temporal_parameters)

    dataset = _depth_subset(dataset, depth_parameters)

    dataset = _update_dataset_coordinate_valid_minmax_attributes(dataset)

    return dataset


def longitude_modulus(longitude: float) -> float:
    """
    Returns the equivalent longitude in [-180, 180[
    """
    # We are using Decimal to avoid issue with rounding
    modulus = float(Decimal(str(longitude + 180)) % 360)
    # Modulus with python return a negative value if the denominator is negative
    # To counteract that, we add 360 if the result is < 0
    modulus = modulus if modulus >= 0 else modulus + 360
    return modulus - 180


def longitude_modulus_upper_bound(longitude: float) -> float:
    """
    Returns the equivalent longitude in ]-180, 180]
    """
    modulus = longitude_modulus(longitude)
    if modulus == -180:
        return 180.0
    return modulus


def check_dataset_subset_bounds(
    username: str,
    password: str,
    dataset_url: str,
    service_type: CopernicusMarineDatasetServiceType,
    dataset_subset: DatasetTimeAndGeographicalSubset,
    subset_method: SubsetMethod,
    dataset_valid_date: Optional[Union[str, int]],
) -> None:
    if service_type in [
        CopernicusMarineDatasetServiceType.GEOSERIES,
        CopernicusMarineDatasetServiceType.TIMESERIES,
        CopernicusMarineDatasetServiceType.OMI_ARCO,
        CopernicusMarineDatasetServiceType.STATIC_ARCO,
    ]:
        dataset = custom_open_zarr.open_zarr(
            dataset_url, copernicus_marine_username=username
        )
        dataset_coordinates = dataset.coords
    else:
        raise ServiceNotSupported(service_type)
    for coordinate_label in COORDINATES_LABEL["latitude"]:
        if coordinate_label in dataset.sizes:
            latitudes = dataset_coordinates[coordinate_label].values
            user_minimum_coordinate_value = (
                dataset_subset.minimum_latitude
                if dataset_subset.minimum_latitude is not None
                else latitudes.min()
            )
            user_maximum_coordinate_value = (
                dataset_subset.maximum_latitude
                if dataset_subset.maximum_latitude is not None
                else latitudes.max()
            )
            _check_coordinate_overlap(
                dimension="latitude",
                user_minimum_coordinate_value=user_minimum_coordinate_value,
                user_maximum_coordinate_value=user_maximum_coordinate_value,
                dataset_minimum_coordinate_value=latitudes.min(),
                dataset_maximum_coordinate_value=latitudes.max(),
                is_strict=subset_method == "strict",
            )
    for coordinate_label in COORDINATES_LABEL["longitude"]:
        if coordinate_label in dataset.sizes:
            longitudes = dataset_coordinates[coordinate_label].values
            _check_coordinate_overlap(
                dimension="longitude",
                user_minimum_coordinate_value=(
                    longitude_modulus(dataset_subset.minimum_longitude)
                    if dataset_subset.minimum_longitude is not None
                    else longitudes.min()
                ),
                user_maximum_coordinate_value=(
                    longitude_modulus_upper_bound(
                        dataset_subset.maximum_longitude
                    )
                    if dataset_subset.maximum_longitude is not None
                    else longitudes.max()
                ),
                dataset_minimum_coordinate_value=longitudes.min(),
                dataset_maximum_coordinate_value=longitudes.max(),
                is_strict=subset_method == "strict",
            )
    for coordinate_label in COORDINATES_LABEL["time"]:
        if coordinate_label in dataset.sizes:
            times = dataset_coordinates[coordinate_label].values
            if dataset_valid_date:
                times_min = dataset_valid_date
            else:
                times_min = times.min()
            dataset_minimum_coordinate_value = date_to_datetime(times_min)
            dataset_maximum_coordinate_value = date_to_datetime(times.max())
            user_minimum_coordinate_value = (
                dataset_subset.start_datetime
                if dataset_subset.start_datetime is not None
                else dataset_minimum_coordinate_value
            )
            user_maximum_coordinate_value = (
                dataset_subset.end_datetime
                if dataset_subset.end_datetime is not None
                else dataset_maximum_coordinate_value
            )
            _check_coordinate_overlap(
                dimension="time",
                user_minimum_coordinate_value=user_minimum_coordinate_value,
                user_maximum_coordinate_value=user_maximum_coordinate_value,
                dataset_minimum_coordinate_value=dataset_minimum_coordinate_value,
                dataset_maximum_coordinate_value=dataset_maximum_coordinate_value,
                is_strict=subset_method == "strict",
            )


def date_to_datetime(date: Union[str, int]) -> datetime:
    if isinstance(date, int):
        return Timestamp(date * 1e6).to_pydatetime()
    else:
        return Timestamp(date).to_pydatetime().replace(tzinfo=None)


@typing.no_type_check
def _check_coordinate_overlap(
    dimension: str,
    user_minimum_coordinate_value: Union[float, datetime],
    user_maximum_coordinate_value: Union[float, datetime],
    dataset_minimum_coordinate_value: Union[float, datetime],
    dataset_maximum_coordinate_value: Union[float, datetime],
    is_strict: bool,
) -> None:
    message = (
        f"Some or all of your subset selection "
        f"[{user_minimum_coordinate_value}, {user_maximum_coordinate_value}] "
        f"for the {dimension} dimension  exceed the dataset coordinates "
        f"[{dataset_minimum_coordinate_value}, "
        f"{dataset_maximum_coordinate_value}]"
    )
    if dimension == "longitude":
        if dataset_minimum_coordinate_value == -180:
            dataset_maximum_coordinate_value = 180
        if dataset_maximum_coordinate_value == 180:
            dataset_minimum_coordinate_value = -180
    if (
        user_maximum_coordinate_value < dataset_minimum_coordinate_value
        or user_minimum_coordinate_value > dataset_maximum_coordinate_value
    ):
        raise CoordinatesOutOfDatasetBounds(message)
    elif (
        (
            user_minimum_coordinate_value < dataset_minimum_coordinate_value
            and user_maximum_coordinate_value
            > dataset_maximum_coordinate_value
        )
        or user_minimum_coordinate_value < dataset_minimum_coordinate_value
        or user_maximum_coordinate_value > dataset_maximum_coordinate_value
    ):
        if is_strict:
            raise CoordinatesOutOfDatasetBounds(message)
        else:
            logger.warning(message)



import functools
import inspect
import logging
from typing import Any, Callable, Dict

import click

from copernicusmarine.core_functions.deprecated_options import (
    DEPRECATED_OPTIONS,
    DeprecatedOptionMapping,
)

logger = logging.getLogger("copernicus_marine_root_logger")


def get_deprecated_message(
    old_value,
    preferred_value,
    deleted_for_v2: bool = False,
    deprecated_for_v2: bool = False,
    only_for_v2: bool = False,
):
    message = ""
    if only_for_v2:
        message = f"Deprecation warning for option '{old_value}'. "
    else:
        message = f"'{old_value}' has been deprecated. "
    if old_value != preferred_value and not only_for_v2:
        message += f"Use '{preferred_value}' instead. "
    if deleted_for_v2:
        message += (
            "This option will no longer be "
            + "available in copernicusmarine>=2.0.0. "
            + "Please refer to the documentation when the new major "
            + "version is released for more information."
        )
    if deprecated_for_v2:
        message += (
            "This option will be deprecated in copernicusmarine>=2.0.0 i.e. "
            + "it will not break but it might have an unexpected effect."
        )
    return message


def log_deprecated_message(
    old_value,
    preferred_value,
    deleted_for_v2: bool,
    deprecated_for_v2: bool,
    only_for_v2: bool,
):
    logger.warning(
        get_deprecated_message(
            old_value,
            preferred_value,
            deleted_for_v2=deleted_for_v2,
            deprecated_for_v2=deprecated_for_v2,
            only_for_v2=only_for_v2,
        )
    )


def raise_both_old_and_new_value_error(old_value, new_value):
    raise TypeError(
        f"Received both {old_value} and {new_value} as arguments! "
        f"{get_deprecated_message(old_value, new_value)}"
    )


class DeprecatedClickOption(click.Option):
    def __init__(self, *args, **kwargs):
        self.deprecated = kwargs.pop("deprecated", ())
        self.preferred = kwargs.pop("preferred", args[0][-1])
        super().__init__(*args, **kwargs)


class DeprecatedClickOptionsCommand(click.Command):
    def make_parser(self, ctx):
        parser = super().make_parser(ctx)

        # get the parser options
        options = set(parser._short_opt.values())
        options |= set(parser._long_opt.values())

        # get name of the command
        command_name = ctx.command.name

        for option in options:

            def make_process(an_option):
                orig_process = an_option.process
                deprecated = getattr(an_option.obj, "deprecated", [])
                preferred = getattr(an_option.obj, "preferred", [])

                def process(value, state):
                    frame = inspect.currentframe()
                    try:
                        opt = frame.f_back.f_locals.get("opt")
                    finally:
                        del frame
                    old_alias = opt.replace("--", "").replace("-", "_")  # type: ignore
                    if (
                        opt in deprecated
                        or old_alias
                        in DEPRECATED_OPTIONS.deprecated_options_by_old_names
                    ):
                        alias_info = (
                            DEPRECATED_OPTIONS.deprecated_options_by_old_names[
                                old_alias
                            ]
                        )
                        if command_name in alias_info.targeted_functions:
                            log_deprecated_message(
                                opt,
                                preferred,
                                alias_info.deleted_for_v2,
                                alias_info.deprecated_for_v2,
                                alias_info.only_for_v2,
                            )
                    return orig_process(value, state)

                return process

            option.process = make_process(option)

        return parser


def deprecated_python_option(
    deprecated_option: DeprecatedOptionMapping,
) -> Callable:
    def deco(f: Callable):
        @functools.wraps(f)
        def wrapper(*args, **kwargs):
            rename_kwargs(f.__name__, kwargs, deprecated_option)
            return f(*args, **kwargs)

        return wrapper

    return deco


def rename_kwargs(
    func_name: str, kwargs: Dict[str, Any], aliases: DeprecatedOptionMapping
):
    for old, alias_info in aliases.deprecated_options_by_old_names.items():
        if func_name not in alias_info.targeted_functions:
            continue
        new = alias_info.new_name
        if old in kwargs:
            if new in kwargs and old != new:
                raise_both_old_and_new_value_error(old, new)
            log_deprecated_message(
                old,
                new,
                alias_info.deleted_for_v2,
                alias_info.deprecated_for_v2,
                alias_info.only_for_v2,
            )
            if alias_info.replace:
                kwargs[new] = kwargs.pop(old)



"""
.
"""

import logging.config
import time
from importlib.metadata import version

from copernicusmarine.logging_conf import logging_configuration_dict

__version__ = version("copernicusmarine")

logging.config.dictConfig(logging_configuration_dict)
logging.Formatter.converter = time.gmtime

from copernicusmarine.python_interface.describe import describe
from copernicusmarine.python_interface.get import get
from copernicusmarine.python_interface.login import login
from copernicusmarine.python_interface.open_dataset import (
    load_xarray_dataset,  # depracated
    open_dataset,
)
from copernicusmarine.python_interface.read_dataframe import (
    load_pandas_dataframe,  # depracated
    read_dataframe,
)
from copernicusmarine.python_interface.subset import subset



from .retry import BeforeRetry  # noqa: F401
from .retry import RetryInfo  # noqa: F401
from .retry import RetryPolicy  # noqa: F401
from .retry import RetryPolicyStrategy  # noqa: F401
from .retry import retry  # noqa: F401



import pathlib
from datetime import datetime
from typing import List, Optional, Union

import pandas

from copernicusmarine.catalogue_parser.request_structure import LoadRequest
from copernicusmarine.core_functions.deprecated import (
    deprecated_python_option,
    log_deprecated_message,
)
from copernicusmarine.core_functions.deprecated_options import (
    DEPRECATED_OPTIONS,
)
from copernicusmarine.core_functions.models import (
    DEFAULT_SUBSET_METHOD,
    SubsetMethod,
)
from copernicusmarine.download_functions.download_arco_series import (
    read_dataframe_from_arco_series,
)
from copernicusmarine.download_functions.subset_parameters import (
    DepthParameters,
    GeographicalParameters,
    LatitudeParameters,
    LongitudeParameters,
    TemporalParameters,
)
from copernicusmarine.python_interface.exception_handler import (
    log_exception_and_exit,
)
from copernicusmarine.python_interface.load_utils import (
    load_data_object_from_load_request,
)
from copernicusmarine.python_interface.utils import homogenize_datetime


@log_exception_and_exit
def load_pandas_dataframe(*args, **kwargs):
    """
    Deprecated function, use 'read_dataframe' instead.
    """
    log_deprecated_message(
        "load_pandas_dataframe",
        "read_dataframe",
        deleted_for_v2=True,
        deprecated_for_v2=False,
        only_for_v2=False,
    )
    return read_dataframe(*args, **kwargs)


@deprecated_python_option(DEPRECATED_OPTIONS)
@log_exception_and_exit
def read_dataframe(
    dataset_url: Optional[str] = None,
    dataset_id: Optional[str] = None,
    dataset_version: Optional[str] = None,
    dataset_part: Optional[str] = None,
    username: Optional[str] = None,
    password: Optional[str] = None,
    variables: Optional[List[str]] = None,
    minimum_longitude: Optional[float] = None,
    maximum_longitude: Optional[float] = None,
    minimum_latitude: Optional[float] = None,
    maximum_latitude: Optional[float] = None,
    minimum_depth: Optional[float] = None,
    maximum_depth: Optional[float] = None,
    vertical_dimension_as_originally_produced: bool = True,
    start_datetime: Optional[Union[datetime, str]] = None,
    end_datetime: Optional[Union[datetime, str]] = None,
    subset_method: SubsetMethod = DEFAULT_SUBSET_METHOD,
    force_service: Optional[str] = None,
    credentials_file: Optional[Union[pathlib.Path, str]] = None,
    overwrite_metadata_cache: bool = False,
    no_metadata_cache: bool = False,
    disable_progress_bar: bool = False,
) -> pandas.DataFrame:
    """
    Immediately loads a Pandas DataFrame into memory from a specified dataset.
    Unlike “lazy-loading”, the data is loaded as soon as this function is executed, which may be preferable when rapid access to the entire data set is required, but may require careful memory management.

    Args:
        dataset_url (str, optional): The URL of the dataset.
        dataset_id (str, optional): The identifier of the dataset.
        dataset_version (str, optional): Force a specific dataset version.
        dataset_part (str, optional): Force a specific dataset part.
        username (str, optional): Username for authentication.
        password (str, optional): Password for authentication.
        variables (List[str], optional): List of variable names to load.
        minimum_longitude (float, optional): Minimum longitude for spatial subset.
        maximum_longitude (float, optional): Maximum longitude for spatial subset.
        minimum_latitude (float, optional): Minimum latitude for spatial subset.
        maximum_latitude (float, optional): Maximum latitude for spatial subset.
        minimum_depth (float, optional): Minimum depth for vertical subset.
        maximum_depth (float, optional): Maximum depth for vertical subset.
        vertical_dimension_as_originally_produced (bool, optional): If True, use the vertical dimension as originally produced.
        start_datetime (datetime, optional): Start datetime for temporal subset.
        end_datetime (datetime, optional): End datetime for temporal subset.
        subset_method (str, optional): The subset method ('nearest' or 'strict') when requesting the dataset. If strict, you can only request dimension strictly inside the dataset.
        force_service (str, optional): Force a specific service for data download.
        credentials_file (Union[pathlib.Path, str], optional): Path to a credentials file for authentication.
        overwrite_metadata_cache (bool, optional): If True, overwrite the metadata cache.
        no_metadata_cache (bool, optional): If True, do not use metadata caching.

    Returns:
        pandas.DataFrame: A DataFrame containing the loaded Copernicus Marine data.
    """  # noqa
    start_datetime = homogenize_datetime(start_datetime)
    end_datetime = homogenize_datetime(end_datetime)
    credentials_file = (
        pathlib.Path(credentials_file) if credentials_file else None
    )
    load_request = LoadRequest(
        dataset_url=dataset_url,
        dataset_id=dataset_id,
        force_dataset_version=dataset_version,
        force_dataset_part=dataset_part,
        username=username,
        password=password,
        variables=variables,
        geographical_parameters=GeographicalParameters(
            latitude_parameters=LatitudeParameters(
                minimum_latitude=minimum_latitude,
                maximum_latitude=maximum_latitude,
            ),
            longitude_parameters=LongitudeParameters(
                minimum_longitude=minimum_longitude,
                maximum_longitude=maximum_longitude,
            ),
        ),
        temporal_parameters=TemporalParameters(
            start_datetime=start_datetime,
            end_datetime=end_datetime,
        ),
        subset_method=subset_method,
        depth_parameters=DepthParameters(
            minimum_depth=minimum_depth,
            maximum_depth=maximum_depth,
            vertical_dimension_as_originally_produced=vertical_dimension_as_originally_produced,  # noqa
        ),
        force_service=force_service,
        credentials_file=credentials_file,
        overwrite_metadata_cache=overwrite_metadata_cache,
        no_metadata_cache=no_metadata_cache,
    )
    dataset = load_data_object_from_load_request(
        load_request,
        disable_progress_bar,
        read_dataframe_from_arco_series,
    )
    return dataset



import logging
import pathlib
from typing import Hashable, Iterable, Literal, Optional, Union

import click
import pandas
import xarray

from copernicusmarine.catalogue_parser.request_structure import SubsetRequest
from copernicusmarine.core_functions import custom_open_zarr
from copernicusmarine.core_functions.utils import (
    FORCE_DOWNLOAD_CLI_PROMPT_MESSAGE,
    add_copernicusmarine_version_in_dataset_attributes,
    get_unique_filename,
)
from copernicusmarine.download_functions.common_download import (
    download_delayed_dataset,
    get_delayed_download,
)
from copernicusmarine.download_functions.subset_parameters import (
    DepthParameters,
    GeographicalParameters,
    LatitudeParameters,
    LongitudeParameters,
    TemporalParameters,
)
from copernicusmarine.download_functions.subset_xarray import (
    date_to_datetime,
    subset,
)
from copernicusmarine.download_functions.utils import (
    FileFormat,
    get_filename,
    get_formatted_dataset_size_estimation,
)

logger = logging.getLogger("copernicus_marine_root_logger")


def _rechunk(dataset: xarray.Dataset) -> xarray.Dataset:
    preferred_chunks = {}
    for variable in dataset:
        preferred_chunks = dataset[variable].encoding["preferred_chunks"]
        del dataset[variable].encoding["chunks"]

    if "depth" in preferred_chunks:
        preferred_chunks["elevation"] = preferred_chunks["depth"]
    elif "elevation" in preferred_chunks:
        preferred_chunks["depth"] = preferred_chunks["elevation"]

    return dataset.chunk(
        _filter_dimensions(preferred_chunks, dataset.sizes.keys())
    )


def _filter_dimensions(
    rechunks: dict[str, int], dimensions: Iterable[Hashable]
) -> dict[str, int]:
    return {k: v for k, v in rechunks.items() if k in dimensions}


def download_dataset(
    username: str,
    password: str,
    dataset_id: str,
    geographical_parameters: GeographicalParameters,
    temporal_parameters: TemporalParameters,
    depth_parameters: DepthParameters,
    dataset_url: str,
    output_directory: pathlib.Path,
    output_filename: Optional[str],
    file_format: FileFormat,
    variables: Optional[list[str]],
    disable_progress_bar: bool,
    netcdf_compression_enabled: bool,
    netcdf_compression_level: Optional[int],
    netcdf3_compatible: bool,
    force_download: bool = False,
    overwrite_output_data: bool = False,
):
    dataset = _rechunk(
        open_dataset_from_arco_series(
            username=username,
            password=password,
            dataset_url=dataset_url,
            variables=variables,
            geographical_parameters=geographical_parameters,
            temporal_parameters=temporal_parameters,
            depth_parameters=depth_parameters,
            chunks="auto",
        )
    )

    dataset = add_copernicusmarine_version_in_dataset_attributes(dataset)

    filename = get_filename(output_filename, dataset, dataset_id, file_format)
    output_path = pathlib.Path(output_directory, filename)
    if not output_directory.is_dir():
        pathlib.Path.mkdir(output_directory, parents=True)

    if not force_download:
        logger.info(dataset)
        logger.info(
            "Estimated size of the dataset file is "
            f"{get_formatted_dataset_size_estimation(dataset)}."
        )
        click.confirm(
            FORCE_DOWNLOAD_CLI_PROMPT_MESSAGE,
            default=True,
            abort=True,
            err=True,
        )
    else:
        logger.info(
            "Estimated size of the dataset file is "
            f"{get_formatted_dataset_size_estimation(dataset)}."
        )
    logger.info("Writing to local storage. Please wait...")

    output_path = get_unique_filename(
        filepath=output_path, overwrite_option=overwrite_output_data
    )

    delayed = get_delayed_download(
        dataset,
        output_path,
        netcdf_compression_enabled,
        netcdf_compression_level,
        netcdf3_compatible,
    )
    download_delayed_dataset(delayed, disable_progress_bar)
    logger.info(f"Successfully downloaded to {output_path}")

    return output_path


def download_zarr(
    username: str,
    password: str,
    subset_request: SubsetRequest,
    dataset_id: str,
    disable_progress_bar: bool,
    dataset_valid_start_date: Optional[Union[str, int]],
):
    geographical_parameters = GeographicalParameters(
        latitude_parameters=LatitudeParameters(
            minimum_latitude=subset_request.minimum_latitude,
            maximum_latitude=subset_request.maximum_latitude,
        ),
        longitude_parameters=LongitudeParameters(
            minimum_longitude=subset_request.minimum_longitude,
            maximum_longitude=subset_request.maximum_longitude,
        ),
    )
    start_datetime = subset_request.start_datetime
    if dataset_valid_start_date:
        minimum_start_date = date_to_datetime(dataset_valid_start_date)
        if (
            not subset_request.start_datetime
            or subset_request.start_datetime < minimum_start_date
        ):
            start_datetime = minimum_start_date

    temporal_parameters = TemporalParameters(
        start_datetime=start_datetime,
        end_datetime=subset_request.end_datetime,
    )
    depth_parameters = DepthParameters(
        minimum_depth=subset_request.minimum_depth,
        maximum_depth=subset_request.maximum_depth,
        vertical_dimension_as_originally_produced=subset_request.vertical_dimension_as_originally_produced,  # noqa
    )
    dataset_url = str(subset_request.dataset_url)
    output_directory = (
        subset_request.output_directory
        if subset_request.output_directory
        else pathlib.Path(".")
    )
    variables = subset_request.variables
    force_download = subset_request.force_download

    output_path = download_dataset(
        username=username,
        password=password,
        dataset_id=dataset_id,
        geographical_parameters=geographical_parameters,
        temporal_parameters=temporal_parameters,
        depth_parameters=depth_parameters,
        dataset_url=dataset_url,
        output_directory=output_directory,
        output_filename=subset_request.output_filename,
        file_format=subset_request.file_format,
        variables=variables,
        disable_progress_bar=disable_progress_bar,
        force_download=force_download,
        overwrite_output_data=subset_request.overwrite_output_data,
        netcdf_compression_enabled=subset_request.netcdf_compression_enabled,
        netcdf_compression_level=subset_request.netcdf_compression_level,
        netcdf3_compatible=subset_request.netcdf3_compatible,
    )
    return output_path


def open_dataset_from_arco_series(
    username: str,
    password: str,
    dataset_url: str,
    variables: Optional[list[str]],
    geographical_parameters: GeographicalParameters,
    temporal_parameters: TemporalParameters,
    depth_parameters: DepthParameters,
    chunks=Optional[Literal["auto"]],
) -> xarray.Dataset:
    dataset = custom_open_zarr.open_zarr(
        dataset_url,
        chunks=chunks,
        copernicus_marine_username=username,
    )
    dataset = subset(
        dataset=dataset,
        variables=variables,
        geographical_parameters=geographical_parameters,
        temporal_parameters=temporal_parameters,
        depth_parameters=depth_parameters,
    )
    return dataset


def read_dataframe_from_arco_series(
    username: str,
    password: str,
    dataset_url: str,
    variables: Optional[list[str]],
    geographical_parameters: GeographicalParameters,
    temporal_parameters: TemporalParameters,
    depth_parameters: DepthParameters,
    chunks: Optional[Literal["auto"]],
) -> pandas.DataFrame:
    dataset = open_dataset_from_arco_series(
        username=username,
        password=password,
        dataset_url=dataset_url,
        variables=variables,
        geographical_parameters=geographical_parameters,
        temporal_parameters=temporal_parameters,
        depth_parameters=depth_parameters,
        chunks=chunks,
    )
    return dataset.to_dataframe()



import json
from typing import Any

from copernicusmarine.core_functions.deprecated import deprecated_python_option
from copernicusmarine.core_functions.deprecated_options import (
    DEPRECATED_OPTIONS,
)
from copernicusmarine.core_functions.describe import describe_function
from copernicusmarine.python_interface.exception_handler import (
    log_exception_and_exit,
)


@deprecated_python_option(DEPRECATED_OPTIONS)
@log_exception_and_exit
def describe(
    include_description: bool = False,
    include_datasets: bool = False,
    include_keywords: bool = False,
    include_versions: bool = False,
    include_all: bool = False,
    contains: list[str] = [],
    overwrite_metadata_cache: bool = False,
    no_metadata_cache: bool = False,
    disable_progress_bar: bool = False,
    staging: bool = False,
) -> dict[str, Any]:
    """
    Retrieve metadata information from the Copernicus Marine catalogue.

    This function fetches metadata information from the Copernicus Marine catalogue
    based on specified parameters and options.

    Args:
        include_description (bool, optional): Whether to include description for each product. Defaults to False.
        include_datasets (bool, optional): Whether to include dataset information. Defaults to False.
        include_keywords (bool, optional): Whether to include keywords for each product. Defaults to False.
        include_versions (bool, optional): Whether to include all versions of each dataset. Defaults to False.
        include_all (bool, optional): Whether to include all metadata information. Defaults to False.
        contains (list[str], optional): List of strings to filter items containing these values. Defaults to [].
        overwrite_metadata_cache (bool, optional): Whether to overwrite the metadata cache. Defaults to False.
        no_metadata_cache (bool, optional): Whether to skip using the metadata cache. Defaults to False.

    Returns:
        dict[str, Any]: A dictionary containing the retrieved metadata information.
    """  # noqa

    if not isinstance(contains, list):
        raise ValueError("contains must be of list type")

    if include_all:
        include_description = True
        include_datasets = True
        include_keywords = True
        include_versions = True

    catalogue_json = describe_function(
        include_description,
        include_datasets,
        include_keywords,
        include_versions,
        contains,
        overwrite_metadata_cache,
        no_metadata_cache,
        disable_progress_bar,
        staging=staging,
    )
    catalogue = json.loads(catalogue_json)
    return catalogue



import pathlib
from datetime import datetime
from typing import List, Optional, Union

import xarray

from copernicusmarine.catalogue_parser.request_structure import LoadRequest
from copernicusmarine.core_functions.deprecated import (
    deprecated_python_option,
    log_deprecated_message,
)
from copernicusmarine.core_functions.deprecated_options import (
    DEPRECATED_OPTIONS,
)
from copernicusmarine.core_functions.models import (
    DEFAULT_SUBSET_METHOD,
    SubsetMethod,
)
from copernicusmarine.download_functions.download_arco_series import (
    open_dataset_from_arco_series,
)
from copernicusmarine.download_functions.subset_parameters import (
    DepthParameters,
    GeographicalParameters,
    LatitudeParameters,
    LongitudeParameters,
    TemporalParameters,
)
from copernicusmarine.python_interface.exception_handler import (
    log_exception_and_exit,
)
from copernicusmarine.python_interface.load_utils import (
    load_data_object_from_load_request,
)
from copernicusmarine.python_interface.utils import homogenize_datetime


@log_exception_and_exit
def load_xarray_dataset(*args, **kwargs):
    """
    Deprecated function, use 'open_dataset' instead.
    """
    log_deprecated_message(
        "load_xarray_dataset",
        "open_dataset",
        deleted_for_v2=True,
        deprecated_for_v2=False,
        only_for_v2=False,
    )
    return open_dataset(*args, **kwargs)


@deprecated_python_option(DEPRECATED_OPTIONS)
@log_exception_and_exit
def open_dataset(
    dataset_url: Optional[str] = None,
    dataset_id: Optional[str] = None,
    dataset_version: Optional[str] = None,
    dataset_part: Optional[str] = None,
    username: Optional[str] = None,
    password: Optional[str] = None,
    variables: Optional[List[str]] = None,
    minimum_longitude: Optional[float] = None,
    maximum_longitude: Optional[float] = None,
    minimum_latitude: Optional[float] = None,
    maximum_latitude: Optional[float] = None,
    minimum_depth: Optional[float] = None,
    maximum_depth: Optional[float] = None,
    vertical_dimension_as_originally_produced: bool = True,
    start_datetime: Optional[Union[datetime, str]] = None,
    end_datetime: Optional[Union[datetime, str]] = None,
    subset_method: SubsetMethod = DEFAULT_SUBSET_METHOD,
    service: Optional[str] = None,
    credentials_file: Optional[Union[pathlib.Path, str]] = None,
    overwrite_metadata_cache: bool = False,
    no_metadata_cache: bool = False,
    disable_progress_bar: bool = False,
) -> xarray.Dataset:
    """
    Load an xarray dataset using "lazy-loading" mode from a Copernicus Marine data source using either the ARCO series protocol.
    This means that data is only loaded into memory when a computation is called, optimizing RAM usage by avoiding immediate loading.
    It supports various parameters for customization, such as specifying ge ographical bounds, temporal range, depth range, and more.

    Args:
        dataset_url (str, optional): The URL of the dataset. Either `dataset_url` or `dataset_id` should be provided.
        dataset_id (str, optional): The ID of the dataset. Either `dataset_url` or `dataset_id` should be provided.
        dataset_version (str, optional): Force the use of a specific dataset version.
        dataset_part (str, optional): Force the use of a specific dataset part.
        username (str, optional): Username for authentication, if required.
        password (str, optional): Password for authentication, if required.
        variables (List[str], optional): List of variable names to be loaded from the dataset.
        minimum_longitude (float, optional): The minimum longitude for subsetting the data.
        maximum_longitude (float, optional): The maximum longitude for subsetting the data.
        minimum_latitude (float, optional): The minimum latitude for subsetting the data.
        maximum_latitude (float, optional): The maximum latitude for subsetting the data.
        minimum_depth (float, optional): The minimum depth for subsetting the data.
        maximum_depth (float, optional): The maximum depth for subsetting the data.
        subset_method (str, optional): The subset method ('nearest' or 'strict') when requesting the dataset. If strict, you can only request dimension strictly inside the dataset.
        vertical_dimension_as_originally_produced (bool, optional): If True, use the vertical dimension as originally produced.
        start_datetime (datetime, optional): The start datetime for temporal subsetting.
        end_datetime (datetime, optional): The end datetime for temporal subsetting.
        service (str, optional): Force the use of a specific service (ARCO geo series or time series).
        credentials_file (Union[pathlib.Path, str], optional): Path to a file containing authentication credentials.
        overwrite_metadata_cache (bool, optional): If True, overwrite the metadata cache.
        no_metadata_cache (bool, optional): If True, do not use the metadata cache.

    Returns:
        xarray.Dataset: The loaded xarray dataset.
    """  # noqa
    start_datetime = homogenize_datetime(start_datetime)
    end_datetime = homogenize_datetime(end_datetime)
    credentials_file = (
        pathlib.Path(credentials_file) if credentials_file else None
    )
    load_request = LoadRequest(
        dataset_url=dataset_url,
        dataset_id=dataset_id,
        force_dataset_version=dataset_version,
        force_dataset_part=dataset_part,
        username=username,
        password=password,
        variables=variables,
        geographical_parameters=GeographicalParameters(
            latitude_parameters=LatitudeParameters(
                minimum_latitude=minimum_latitude,
                maximum_latitude=maximum_latitude,
            ),
            longitude_parameters=LongitudeParameters(
                minimum_longitude=minimum_longitude,
                maximum_longitude=maximum_longitude,
            ),
        ),
        temporal_parameters=TemporalParameters(
            start_datetime=start_datetime,
            end_datetime=end_datetime,
        ),
        depth_parameters=DepthParameters(
            minimum_depth=minimum_depth,
            maximum_depth=maximum_depth,
            vertical_dimension_as_originally_produced=vertical_dimension_as_originally_produced,  # noqa
        ),
        subset_method=subset_method,
        force_service=service,
        credentials_file=credentials_file,
        overwrite_metadata_cache=overwrite_metadata_cache,
        no_metadata_cache=no_metadata_cache,
    )
    dataset = load_data_object_from_load_request(
        load_request,
        disable_progress_bar,
        open_dataset_from_arco_series,
    )
    return dataset



import logging

import click

from copernicusmarine.command_line_interface.exception_handler import (
    log_exception_and_exit,
)
from copernicusmarine.command_line_interface.utils import (
    MutuallyExclusiveOption,
    tqdm_disable_option,
)
from copernicusmarine.core_functions.deprecated import (
    DeprecatedClickOption,
    DeprecatedClickOptionsCommand,
)
from copernicusmarine.core_functions.describe import describe_function

logger = logging.getLogger("copernicus_marine_root_logger")
blank_logger = logging.getLogger("copernicus_marine_blank_logger")


@click.group()
def cli_group_describe() -> None:
    pass


@cli_group_describe.command(
    "describe",
    cls=DeprecatedClickOptionsCommand,
    short_help="Print Copernicus Marine catalog as JSON.",
    help="""
    Print Copernicus Marine catalog as JSON.

    The default display contains information on the products, and more data
    can be displayed using the --include-<argument> flags.

    The --contains option allows the user to specify one or several strings to
    filter through the catalogue display. The search is performed recursively
    on all attributes of the catalogue, and the tokens only need to be
    contained in one of the attributes (i.e. not exact match).
    """,
    epilog="""
    Examples:

    \b
    copernicusmarine describe --contains METOFFICE-GLO-SST-L4-NRT-OBS-SST-V2 --include-datasets

    \b
    copernicusmarine describe -c METOFFICE-GLO-SST-L4-NRT-OBS-SST-V2
    """,  # noqa
)
@click.option(
    "--include-description",
    type=bool,
    is_flag=True,
    default=False,
    help="Include product description in output.",
)
@click.option(
    "--include-datasets",
    type=bool,
    is_flag=True,
    default=False,
    help="Include product dataset details in output.",
)
@click.option(
    "--include-keywords",
    type=bool,
    is_flag=True,
    default=False,
    help="Include product keyword details in output.",
)
@click.option(
    "--include-versions",
    "--include-all-versions",
    cls=DeprecatedClickOption,
    deprecated=["--include-all-versions"],
    preferred="--include-versions",
    type=bool,
    is_flag=True,
    default=False,
    help="Include dataset versions in output. "
    "By default, shows only the default version.",
)
@click.option(
    "-a",
    "--include-all",
    type=bool,
    is_flag=True,
    default=False,
    help="Include all the possible data in output: "
    "description, datasets, keywords, and versions.",
)
@click.option(
    "--contains",
    "-c",
    type=str,
    multiple=True,
    help="Filter catalogue output. Returns products with attributes "
    "matching a string token.",
)
@click.option(
    "--overwrite-metadata-cache",
    cls=MutuallyExclusiveOption,
    type=bool,
    is_flag=True,
    default=False,
    help="Force to refresh the catalogue by overwriting the local cache.",
    mutually_exclusive=["no_metadata_cache"],
)
@click.option(
    "--no-metadata-cache",
    cls=MutuallyExclusiveOption,
    type=bool,
    is_flag=True,
    default=False,
    help="Bypass the use of cache.",
    mutually_exclusive=["overwrite_metadata_cache"],
)
@tqdm_disable_option
@click.option(
    "--log-level",
    type=click.Choice(["DEBUG", "INFO", "WARN", "ERROR", "CRITICAL", "QUIET"]),
    default="INFO",
    help=(
        "Set the details printed to console by the command "
        "(based on standard logging library)."
    ),
)
@click.option(
    "--staging",
    type=bool,
    default=False,
    is_flag=True,
    hidden=True,
)
@log_exception_and_exit
def describe(
    include_description: bool,
    include_datasets: bool,
    include_keywords: bool,
    include_versions: bool,
    include_all: bool,
    contains: list[str],
    overwrite_metadata_cache: bool,
    no_metadata_cache: bool,
    disable_progress_bar: bool,
    log_level: str,
    staging: bool,
) -> None:
    if log_level == "QUIET":
        logger.disabled = True
        logger.setLevel(level="CRITICAL")
    else:
        logger.setLevel(level=log_level)

    if logger.isEnabledFor(logging.DEBUG):
        logger.debug("DEBUG mode activated")

    if include_all:
        include_description = True
        include_datasets = True
        include_keywords = True
        include_versions = True

    json_dump = describe_function(
        include_description=include_description,
        include_datasets=include_datasets,
        include_keywords=include_keywords,
        include_versions=include_versions,
        contains=contains,
        overwrite_metadata_cache=overwrite_metadata_cache,
        no_metadata_cache=no_metadata_cache,
        disable_progress_bar=disable_progress_bar,
        staging=staging,
    )
    blank_logger.info(json_dump)



import datetime
import logging
import os
import pathlib
import re
from itertools import chain
from multiprocessing.pool import ThreadPool
from pathlib import Path
from typing import Iterator, List, Optional, Tuple

import click
from botocore.client import ClientError
from numpy import append, arange
from tqdm import tqdm

from copernicusmarine.catalogue_parser.request_structure import (
    GetRequest,
    overload_regex_with_additionnal_filter,
)
from copernicusmarine.core_functions.environment_variables import (
    COPERNICUSMARINE_GET_CONCURRENT_DOWNLOADS,
)
from copernicusmarine.core_functions.sessions import (
    get_configured_boto3_session,
)
from copernicusmarine.core_functions.utils import (
    FORCE_DOWNLOAD_CLI_PROMPT_MESSAGE,
    flatten,
    get_unique_filename,
    parse_access_dataset_url,
)

logger = logging.getLogger("copernicus_marine_root_logger")
blank_logger = logging.getLogger("copernicus_marine_blank_logger")

NUMBER_THREADS = (
    int(COPERNICUSMARINE_GET_CONCURRENT_DOWNLOADS)
    if COPERNICUSMARINE_GET_CONCURRENT_DOWNLOADS
    else None
)


def download_original_files(
    username: str,
    password: str,
    get_request: GetRequest,
    disable_progress_bar: bool,
    create_file_list: Optional[str],
) -> list[pathlib.Path]:
    files_not_found: list[str] = []
    filenames_in_sync_ignored: list[str] = []
    total_size: float = 0.0
    sizes: list[float] = []
    last_modified_datetimes: list[datetime.datetime] = []
    filenames_in: list[str] = []
    if get_request.direct_download:
        (
            locator,
            filenames_in,
            sizes,
            last_modified_datetimes,
            total_size,
            filenames_in_sync_ignored,
            files_not_found,
        ) = _download_header_for_direct_download(
            get_request.direct_download,
            str(get_request.dataset_url),
            get_request.sync,
            pathlib.Path(get_request.output_directory),
            username,
        )
    if not get_request.direct_download or files_not_found or get_request.regex:
        if files_not_found:
            files_not_found_regex = "|".join(
                [
                    re.escape(file_not_found)
                    for file_not_found in files_not_found
                ]
            )
            get_request.regex = overload_regex_with_additionnal_filter(
                files_not_found_regex, get_request.regex
            )
        result = _download_header(
            str(get_request.dataset_url),
            get_request.regex,
            username,
            password,
            get_request.sync,
            create_file_list,
            pathlib.Path(get_request.output_directory),
            only_list_root_path=get_request.index_parts,
            overwrite=get_request.overwrite_output_data,
        )
        if result:
            (
                locator,
                filenames_in_listing,
                sizes_listing,
                last_modified_datetimes_listing,
                total_size_listing,
                filenames_in_sync_ignored_listing,
            ) = result
            filenames_in.extend(filenames_in_listing)
            filenames_in_sync_ignored.extend(filenames_in_sync_ignored_listing)
            total_size += total_size_listing
            sizes.extend(sizes_listing)
            last_modified_datetimes.extend(last_modified_datetimes_listing)
        elif not get_request.direct_download or len(files_not_found) == len(
            get_request.direct_download
        ):
            return []
    message = _create_information_message_before_download(
        filenames_in, sizes, last_modified_datetimes, total_size
    )
    filenames_out = create_filenames_out(
        filenames_in=filenames_in,
        output_directory=pathlib.Path(get_request.output_directory),
        no_directories=get_request.no_directories,
        overwrite=(
            get_request.overwrite_output_data
            if not get_request.sync
            else False
        ),
    )
    if not get_request.force_download and total_size:
        logger.info(message)
    if get_request.show_outputnames:
        for filename_out in filenames_out:
            blank_logger.info(filename_out)
    files_to_delete = []
    if get_request.sync_delete:
        filenames_out_sync_ignored = create_filenames_out(
            filenames_in=filenames_in_sync_ignored,
            output_directory=pathlib.Path(get_request.output_directory),
            no_directories=get_request.no_directories,
            overwrite=False,
            unique_names_compared_to_local_files=False,
        )
        files_to_delete = _get_files_to_delete_with_sync(
            filenames_in=filenames_in_sync_ignored,
            output_directory=pathlib.Path(get_request.output_directory),
            filenames_out=filenames_out_sync_ignored,
        )
        if files_to_delete:
            logger.info("Some files will be deleted due to sync delete:")
            for file_to_delete in files_to_delete:
                logger.info(file_to_delete)
    if not total_size:
        logger.info("No data to download")
        if not files_to_delete:
            return []
    if not get_request.force_download:
        click.confirm(
            FORCE_DOWNLOAD_CLI_PROMPT_MESSAGE,
            default=True,
            abort=True,
            err=True,
        )
    endpoint: str
    bucket: str
    endpoint, bucket = locator
    if get_request.sync_delete and files_to_delete:
        for file_to_delete in files_to_delete:
            file_to_delete.unlink()
    return download_files(
        username,
        endpoint,
        bucket,
        filenames_in,
        filenames_out,
        disable_progress_bar,
    )


def _get_files_to_delete_with_sync(
    filenames_in: list[str],
    output_directory: pathlib.Path,
    filenames_out: list[Path],
) -> list[pathlib.Path]:
    product_structure = str(
        _local_path_from_s3_url(filenames_in[0], Path(""))
    ).split("/")
    product_id = product_structure[0]
    dataset_id = product_structure[1]
    dataset_level_local_folder = output_directory / product_id / dataset_id
    files_to_delete = []
    for local_file in dataset_level_local_folder.glob("**/*"):
        if local_file.is_file() and local_file not in filenames_out:
            files_to_delete.append(local_file)
    return files_to_delete


def download_files(
    username: str,
    endpoint_url: str,
    bucket: str,
    filenames_in: List[str],
    filenames_out: List[pathlib.Path],
    disable_progress_bar: bool,
) -> list[pathlib.Path]:
    nfiles_per_process, nfiles = 1, len(filenames_in)
    indexes = append(
        arange(0, nfiles, nfiles_per_process, dtype=int),
        nfiles,
    )
    groups_in_files = [
        filenames_in[indexes[i] : indexes[i + 1]]
        for i in range(len(indexes) - 1)
    ]
    groups_out_files = [
        filenames_out[indexes[i] : indexes[i + 1]]
        for i in range(len(indexes) - 1)
    ]

    for groups_out_file in groups_out_files:
        parent_dir = Path(groups_out_file[0]).parent
        if not parent_dir.is_dir():
            pathlib.Path.mkdir(parent_dir, parents=True)

    # TODO: v2 It would be proably better to use an async approach
    # TODO: v2 probably better to use an argument for the number
    # of threads instead of using the environment variable
    if NUMBER_THREADS is None or NUMBER_THREADS:
        pool = ThreadPool(processes=NUMBER_THREADS)
        download_summary_list: Iterator[List[Path]] = pool.imap(
            _download_files,
            zip(
                [username] * len(groups_in_files),
                [endpoint_url] * len(groups_in_files),
                [bucket] * len(groups_in_files),
                groups_in_files,
                groups_out_files,
            ),
        )
    else:
        logger.info("Downloading files one by one...")
        download_summary_list = map(
            _download_files,
            zip(
                [username] * len(groups_in_files),
                [endpoint_url] * len(groups_in_files),
                [bucket] * len(groups_in_files),
                groups_in_files,
                groups_out_files,
            ),
        )
    download_summary = list(
        tqdm(
            download_summary_list,
            total=len(groups_in_files),
            disable=disable_progress_bar,
        )
    )
    return flatten(download_summary)


def _download_header(
    data_path: str,
    regex: Optional[str],
    username: str,
    _password: str,
    sync: bool,
    create_file_list: Optional[str],
    directory_out: pathlib.Path,
    only_list_root_path: bool = False,
    overwrite: bool = False,
) -> Optional[
    Tuple[
        Tuple[str, str],
        list[str],
        List[float],
        List[datetime.datetime],
        float,
        list[str],
    ]
]:
    (endpoint_url, bucket, path) = parse_access_dataset_url(
        data_path, only_dataset_root_path=only_list_root_path
    )

    filenames: list[str] = []
    sizes: list[float] = []
    total_size = 0.0
    last_modified_datetimes: list[datetime.datetime] = []
    etags: list[str] = []
    raw_filenames = _list_files_on_marine_data_lake_s3(
        username, endpoint_url, bucket, path, not only_list_root_path
    )
    filenames_without_sync = []
    for filename, size, last_modified_datetime, etag in raw_filenames:
        if not regex or re.search(regex, filename):
            filenames_without_sync.append(filename)
            if not sync or _check_needs_to_be_synced(
                filename, size, last_modified_datetime, directory_out
            ):
                filenames.append(filename)
                sizes.append(float(size))
                last_modified_datetimes.append(last_modified_datetime)
                etags.append(etag)
    total_size = sum(sizes)
    if create_file_list and create_file_list.endswith(".txt"):
        download_filename = get_unique_filename(
            directory_out / create_file_list, overwrite
        )
        logger.info(f"The file list is written at {download_filename}")
        with open(download_filename, "w") as file_out:
            for filename in filenames:
                file_out.write(f"{filename}\n")
        return None
    elif create_file_list and create_file_list.endswith(".csv"):
        download_filename = get_unique_filename(
            directory_out / create_file_list, overwrite
        )
        logger.info(f"The file list is written at {download_filename}")
        with open(download_filename, "w") as file_out:
            file_out.write("filename,size,last_modified_datetime,etag\n")
            for (
                filename,
                size_file,
                last_modified_datetime,
                etag,
            ) in zip(filenames, sizes, last_modified_datetimes, etags):
                file_out.write(
                    f"{filename},{size_file},{last_modified_datetime},{etag}\n"
                )
        return None
    locator = (endpoint_url, bucket)
    return (
        locator,
        filenames,
        sizes,
        last_modified_datetimes,
        total_size,
        filenames_without_sync,
    )


def _download_header_for_direct_download(
    files_to_download: list[str],
    dataset_url: str,
    sync: bool,
    directory_out: pathlib.Path,
    username: str,
) -> Tuple[
    Tuple[str, str],
    List[str],
    List[float],
    List[datetime.datetime],
    float,
    list[str],
    list[str],
]:
    (endpoint_url, bucket, path) = parse_access_dataset_url(dataset_url)
    splitted_path = path.split("/")
    root_folder = splitted_path[0]
    product_id = splitted_path[1]
    dataset_id_with_tag = splitted_path[2]

    sizes = []
    last_modified_datetimes = []
    filenames_in = []
    filenames_without_sync = []
    filenames_not_found = []
    for file_to_download in files_to_download:
        file_path = file_to_download.split(f"{dataset_id_with_tag}/")[-1]
        if not file_path:
            logger.warning(
                f"{file_to_download} does not seem to be valid. Skipping."
            )
            filenames_not_found.append(file_path)
            continue
        full_path = (
            f"s3://{bucket}/{root_folder}/{product_id}/"
            f"{dataset_id_with_tag}/{file_path}"
        )
        size_and_last_modified = _get_file_size_and_last_modified(
            endpoint_url, bucket, full_path, username
        )
        if size_and_last_modified:
            size, last_modified = size_and_last_modified
            if not sync or _check_needs_to_be_synced(
                full_path, size, last_modified, directory_out
            ):
                filenames_in.append(full_path)
                sizes.append(float(size))
                last_modified_datetimes.append(last_modified)
            else:
                filenames_without_sync.append(full_path)
        else:
            filenames_not_found.append(file_path)
    if not filenames_in:
        logger.warning(
            "No files found to download for direct download. "
            "Please check the files to download. "
            "We will try to list the files available for download "
            "and compare them with the requested files."
        )
    total_size = sum([size for size in sizes])
    return (
        (endpoint_url, bucket),
        filenames_in,
        sizes,
        last_modified_datetimes,
        total_size,
        filenames_without_sync,
        filenames_not_found,
    )


def _check_needs_to_be_synced(
    filename: str,
    size: int,
    last_modified_datetime: datetime.datetime,
    directory_out: pathlib.Path,
) -> bool:
    filename_out = _local_path_from_s3_url(filename, directory_out)
    if not filename_out.is_file():
        return True
    else:
        file_stats = filename_out.stat()
        if file_stats.st_size != size:
            return True
        else:
            last_created_datetime_out = datetime.datetime.fromtimestamp(
                file_stats.st_ctime, tz=datetime.timezone.utc
            )
            return last_modified_datetime > last_created_datetime_out


def _create_information_message_before_download(
    filenames: list[str],
    sizes: list[float],
    last_modified_datetimes: list[datetime.datetime],
    total_size: float,
) -> str:
    message = "You requested the download of the following files:\n"
    for filename, size, last_modified_datetime in zip(
        filenames[:20], sizes[:20], last_modified_datetimes[:20]
    ):
        message += str(filename)
        datetime_iso = re.sub(
            r"\+00:00$",
            "Z",
            last_modified_datetime.astimezone(datetime.timezone.utc).isoformat(
                timespec="seconds"
            ),
        )
        message += f" - {format_file_size(float(size))} - {datetime_iso}\n"
    if len(filenames) > 20:
        message += f"Printed 20 out of {len(filenames)} files\n"
    message += (
        f"\nTotal size of the download: {format_file_size(total_size)}\n\n"
    )
    return message


def _local_path_from_s3_url(
    s3_url: str, local_directory: pathlib.Path
) -> pathlib.Path:
    return local_directory / pathlib.Path("/".join(s3_url.split("/")[4:]))


def _list_files_on_marine_data_lake_s3(
    username: str,
    endpoint_url: str,
    bucket: str,
    prefix: str,
    recursive: bool,
) -> list[tuple[str, int, datetime.datetime, str]]:

    s3_client, _ = get_configured_boto3_session(
        endpoint_url, ["ListObjects"], username
    )

    paginator = s3_client.get_paginator("list_objects")
    page_iterator = paginator.paginate(
        Bucket=bucket,
        Prefix=prefix,
        Delimiter="/" if not recursive else "",
    )

    s3_objects = chain(
        *map(lambda page: page.get("Contents", []), page_iterator)
    )

    files_already_found: list[tuple[str, int, datetime.datetime, str]] = []
    for s3_object in s3_objects:
        files_already_found.append(
            (
                f"s3://{bucket}/" + s3_object["Key"],
                s3_object["Size"],
                s3_object["LastModified"],
                s3_object["ETag"],
            )
        )
    return files_already_found


def _get_file_size_and_last_modified(
    endpoint_url: str, bucket: str, file_in: str, username: str
) -> Optional[Tuple[int, datetime.datetime]]:
    s3_client, _ = get_configured_boto3_session(
        endpoint_url, ["HeadObject"], username
    )

    try:
        s3_object = s3_client.head_object(
            Bucket=bucket,
            Key=file_in.replace(f"s3://{bucket}/", ""),
        )
        return s3_object["ContentLength"], s3_object["LastModified"]
    except ClientError as e:
        if "404" in str(e):
            logger.warning(
                f"File {file_in} not found on the server. Skipping."
            )
            return None
        else:
            raise e


def _download_files(
    tuple_original_files_filename: Tuple[
        str, str, str, list[str], list[pathlib.Path]
    ],
) -> list[pathlib.Path]:
    (
        username,
        endpoint_url,
        bucket,
        filenames_in,
        filenames_out,
    ) = tuple_original_files_filename

    def _original_files_file_download(
        endpoint_url: str, bucket: str, file_in: str, file_out: pathlib.Path
    ) -> pathlib.Path:
        """
        Download ONE file and return a string of the result
        """
        s3_client, s3_resource = get_configured_boto3_session(
            endpoint_url,
            ["GetObject", "HeadObject"],
            username,
            return_ressources=True,
        )
        last_modified_date_epoch = s3_resource.Object(
            bucket, file_in.replace(f"s3://{bucket}/", "")
        ).last_modified.timestamp()

        s3_client.download_file(
            bucket,
            file_in.replace(f"s3://{bucket}/", ""),
            file_out,
        )

        os.utime(
            file_out, (last_modified_date_epoch, last_modified_date_epoch)
        )

        return file_out

    download_summary = []
    for file_in, file_out in zip(filenames_in, filenames_out):
        download_summary.append(
            _original_files_file_download(
                endpoint_url, bucket, file_in, file_out
            )
        )
    return download_summary


# /////////////////////////////
# --- Tools
# /////////////////////////////


def create_filenames_out(
    filenames_in: list[str],
    overwrite: bool,
    output_directory: pathlib.Path = pathlib.Path("."),
    no_directories=False,
    unique_names_compared_to_local_files=True,
) -> list[pathlib.Path]:
    filenames_out = []
    for filename_in in filenames_in:
        if no_directories:
            filename_out = (
                pathlib.Path(output_directory) / pathlib.Path(filename_in).name
            )
        else:
            # filename_in: s3://mdl-native-xx/native/<product-id>..
            filename_out = _local_path_from_s3_url(
                filename_in, output_directory
            )
        if unique_names_compared_to_local_files:
            filename_out = get_unique_filename(
                filepath=filename_out, overwrite_option=overwrite
            )

        filenames_out.append(filename_out)
    return filenames_out


def format_file_size(
    size: float, decimals: int = 2, binary_system: bool = False
) -> str:
    if binary_system:
        units: list[str] = [
            "B",
            "KiB",
            "MiB",
            "GiB",
            "TiB",
            "PiB",
            "EiB",
            "ZiB",
        ]
        largest_unit: str = "YiB"
        step: int = 1024
    else:
        units = ["B", "kB", "MB", "GB", "TB", "PB", "EB", "ZB"]
        largest_unit = "YB"
        step = 1000

    for unit in units:
        if size < step:
            return ("%." + str(decimals) + "f %s") % (size, unit)
        size /= step

    return ("%." + str(decimals) + "f %s") % (size, largest_unit)



import logging
import os

logger = logging.getLogger("copernicus_marine_root_logger")

COPERNICUSMARINE_SERVICE_USERNAME = os.getenv(
    "COPERNICUSMARINE_SERVICE_USERNAME"
) or os.getenv("COPERNICUS_MARINE_SERVICE_USERNAME")
if os.getenv("COPERNICUS_MARINE_SERVICE_USERNAME"):
    logger.warning(
        "COPERNICUS_MARINE_SERVICE_USERNAME is deprecated. "
        "Please use COPERNICUSMARINE_SERVICE_USERNAME instead."
    )
COPERNICUSMARINE_SERVICE_PASSWORD = os.getenv(
    "COPERNICUSMARINE_SERVICE_PASSWORD"
) or os.getenv("COPERNICUS_MARINE_SERVICE_PASSWORD")
if os.getenv("COPERNICUS_MARINE_SERVICE_PASSWORD"):
    logger.warning(
        "COPERNICUS_MARINE_SERVICE_PASSWORD is deprecated. "
        "Please use COPERNICUSMARINE_SERVICE_PASSWORD instead."
    )

COPERNICUSMARINE_CACHE_DIRECTORY = os.getenv(
    "COPERNICUSMARINE_CACHE_DIRECTORY", ""
)

COPERNICUSMARINE_MAX_CONCURRENT_REQUESTS = os.getenv(
    "COPERNICUSMARINE_MAX_CONCURRENT_REQUESTS", "15"
)

COPERNICUSMARINE_GET_CONCURRENT_DOWNLOADS = os.getenv(
    "COPERNICUSMARINE_GET_CONCURRENT_DOWNLOADS", None
)

COPERNICUSMARINE_DISABLE_SSL_CONTEXT = os.getenv(
    "COPERNICUSMARINE_DISABLE_SSL_CONTEXT"
)
COPERNICUSMARINE_TRUST_ENV = os.getenv("COPERNICUSMARINE_TRUST_ENV", "True")

PROXY_HTTPS = os.getenv("HTTPS_PROXY", "")
PROXY_HTTP = os.getenv("HTTP_PROXY", "")



import base64
import configparser
import logging
import pathlib
from datetime import timedelta
from netrc import netrc
from platform import system
from typing import Literal, Optional, Tuple

import click
import lxml.html
import requests
from cachier.core import cachier

from copernicusmarine.core_functions.environment_variables import (
    COPERNICUSMARINE_SERVICE_PASSWORD,
    COPERNICUSMARINE_SERVICE_USERNAME,
)
from copernicusmarine.core_functions.sessions import (
    get_configured_requests_session,
)
from copernicusmarine.core_functions.utils import (
    CACHE_BASE_DIRECTORY,
    DEFAULT_CLIENT_BASE_DIRECTORY,
)

logger = logging.getLogger("copernicus_marine_root_logger")

DEFAULT_CLIENT_CREDENTIALS_FILENAME = ".copernicusmarine-credentials"
DEFAULT_CLIENT_CREDENTIALS_FILEPATH = (
    DEFAULT_CLIENT_BASE_DIRECTORY / DEFAULT_CLIENT_CREDENTIALS_FILENAME
)


class CredentialCannotBeNone(Exception):
    ...


class InvalidUsernameOrPassword(Exception):
    ...


def _load_credential_from_copernicus_marine_configuration_file(
    credential_type: Literal["username", "password"],
    configuration_filename: pathlib.Path,
) -> Optional[str]:
    configuration_file = open(configuration_filename)
    configuration_string = base64.standard_b64decode(
        configuration_file.read()
    ).decode("utf8")
    config = configparser.RawConfigParser()
    config.read_string(configuration_string)
    credential = config.get("credentials", credential_type)
    if credential:
        logger.debug(f"{credential_type} loaded from {configuration_filename}")
    return credential


def _load_credential_from_netrc_configuration_file(
    credential_type: Literal["username", "password"],
    configuration_filename: pathlib.Path,
    host: str,
) -> Optional[str]:
    authenticator = netrc(configuration_filename).authenticators(host=host)
    if authenticator:
        username, _, password = authenticator
        logger.debug(f"{credential_type} loaded from {configuration_filename}")
        return username if credential_type == "username" else password
    else:
        return None


def _load_credential_from_motu_configuration_file(
    credential_type: Literal["username", "password"],
    configuration_filename: pathlib.Path,
) -> Optional[str]:
    motu_file = open(configuration_filename)
    motu_credential_type = "user" if credential_type == "username" else "pwd"
    config = configparser.RawConfigParser()
    config.read_string(motu_file.read())
    credential = config.get("Main", motu_credential_type)
    if credential:
        logger.debug(f"{credential_type} loaded from {configuration_filename}")
    return credential


def _retrieve_credential_from_prompt(
    credential_type: Literal["username", "password"], hide_input: bool
) -> str:
    return click.prompt(credential_type, hide_input=hide_input)


def _retrieve_credential_from_environment_variable(
    credential_type: Literal["username", "password"]
) -> Optional[str]:
    if credential_type == "username":
        logger.debug("username loaded from environment variable")
        return COPERNICUSMARINE_SERVICE_USERNAME
    if credential_type == "password":
        logger.debug("password loaded from environment variable")
        return COPERNICUSMARINE_SERVICE_PASSWORD


def _retrieve_credential_from_custom_configuration_files(
    credential_type: Literal["username", "password"],
    credentials_file: pathlib.Path,
    host: str = "default_host",
) -> Optional[str]:
    credential = _load_credential_from_copernicus_marine_configuration_file(
        credential_type, credentials_file
    )
    if not credential:
        credential = _load_credential_from_motu_configuration_file(
            credential_type, credentials_file
        )
        if not credential:
            credential = _load_credential_from_netrc_configuration_file(
                credential_type, credentials_file, host=host
            )
    return credential


def _retrieve_credential_from_default_configuration_files(
    credential_type: Literal["username", "password"],
    host: str = "default_host",
) -> Optional[str]:
    copernicus_marine_configuration_file = pathlib.Path(
        DEFAULT_CLIENT_CREDENTIALS_FILEPATH
    )
    motu_configuration_file = pathlib.Path(
        pathlib.Path.home() / "motuclient" / "motuclient-python.ini"
    )
    netrc_configuration_file = pathlib.Path(
        pathlib.Path.home() / ("_netrc" if system() == "Windows" else ".netrc")
    )
    if copernicus_marine_configuration_file.exists():
        credential = (
            _load_credential_from_copernicus_marine_configuration_file(
                credential_type,
                copernicus_marine_configuration_file,
            )
        )
    elif motu_configuration_file.exists():
        credential = _load_credential_from_motu_configuration_file(
            credential_type, motu_configuration_file
        )
    elif netrc_configuration_file.exists():
        credential = _load_credential_from_netrc_configuration_file(
            credential_type, netrc_configuration_file, host=host
        )
    else:
        credential = None
    return credential


def _retrieve_credential_from_configuration_files(
    credential_type: Literal["username", "password"],
    credentials_file: Optional[pathlib.Path],
    host: str = "default_host",
) -> Optional[str]:
    if credentials_file and credentials_file.exists():
        credential = _retrieve_credential_from_custom_configuration_files(
            credential_type, credentials_file, host
        )
    else:
        credential = _retrieve_credential_from_default_configuration_files(
            credential_type, host
        )
    return credential


def copernicusmarine_configuration_file_exists(
    configuration_file_directory: pathlib.Path,
) -> bool:
    configuration_filename = pathlib.Path(
        configuration_file_directory / DEFAULT_CLIENT_CREDENTIALS_FILENAME
    )
    return configuration_filename.exists()


def copernicusmarine_configuration_file_is_valid(
    configuration_file_directory: pathlib.Path,
) -> bool:
    configuration_filename = pathlib.Path(
        configuration_file_directory / DEFAULT_CLIENT_CREDENTIALS_FILENAME
    )
    username = _retrieve_credential_from_configuration_files(
        "username", configuration_filename
    )
    password = _retrieve_credential_from_configuration_files(
        "password", configuration_filename
    )
    return (
        username is not None
        and password is not None
        and _check_credentials_with_cas(username, password)
    )


def create_copernicusmarine_configuration_file(
    username: str,
    password: str,
    configuration_file_directory: pathlib.Path,
    overwrite_configuration_file: bool,
) -> pathlib.Path:
    configuration_lines = [
        "[credentials]\n",
        f"username={username}\n",
        f"password={password}\n",
    ]
    configuration_filename = pathlib.Path(
        configuration_file_directory / DEFAULT_CLIENT_CREDENTIALS_FILENAME
    )
    if configuration_filename.exists() and not overwrite_configuration_file:
        click.confirm(
            f"File {configuration_filename} already exists, overwrite it ?",
            abort=True,
        )
    configuration_file_directory.mkdir(parents=True, exist_ok=True)
    configuration_file = open(configuration_filename, "w")
    configuration_string = base64.b64encode(
        "".join(configuration_lines).encode("ascii", "strict")
    ).decode("utf8")
    configuration_file.write(configuration_string)
    configuration_file.close()
    return configuration_filename


def _check_credentials_with_cas(username: str, password: str) -> bool:
    logger.debug("Checking user credentials...")
    service = "copernicus-marine-client"
    cmems_cas_login_url = (
        f"https://cmems-cas.cls.fr/cas/login?service={service}"
    )
    conn_session = get_configured_requests_session()
    login_session = conn_session.get(
        cmems_cas_login_url, proxies=conn_session.proxies
    )
    login_from_html = lxml.html.fromstring(login_session.text)
    hidden_elements_from_html = login_from_html.xpath(
        '//form//input[@type="hidden"]'
    )
    playload = {
        he.attrib["name"]: he.attrib["value"]
        for he in hidden_elements_from_html
    }
    playload["username"] = username
    playload["password"] = password
    logger.debug(f"POSTing credentials to {cmems_cas_login_url}...")
    login_response = conn_session.post(
        cmems_cas_login_url, data=playload, proxies=conn_session.proxies
    )
    login_success = 'class="success"' in login_response.text
    logger.debug("User credentials checked")
    return login_success


@cachier(stale_after=timedelta(hours=48), cache_dir=CACHE_BASE_DIRECTORY)
def _are_copernicus_marine_credentials_valid(
    username: str, password: str
) -> Optional[bool]:
    number_of_retry = 3
    user_is_active = None  # Not cached by cachier
    while (user_is_active not in [True, False]) and number_of_retry > 0:
        try:
            user_is_active = _check_credentials_with_cas(
                username=username, password=password
            )
        except requests.exceptions.ConnectTimeout:
            number_of_retry -= 1
        except requests.exceptions.ConnectionError:
            number_of_retry -= 1
    return user_is_active


def get_credential(
    credential: Optional[str],
    credential_type: Literal["username", "password"],
    hide_input: bool,
    credentials_file: Optional[pathlib.Path],
) -> str:
    if not credential:
        credential = _retrieve_credential_from_environment_variable(
            credential_type
        )
        if not credential:
            credential = _retrieve_credential_from_configuration_files(
                credential_type=credential_type,
                credentials_file=credentials_file,
                host="nrt.cmems-du.eu",
            )
            if not credential:
                credential = _retrieve_credential_from_configuration_files(
                    credential_type=credential_type,
                    credentials_file=credentials_file,
                    host="my.cmems-du.eu",
                )
                if not credential:
                    credential = _retrieve_credential_from_prompt(
                        credential_type, hide_input=hide_input
                    )
                    if not credential:
                        raise ValueError(f"{credential} cannot be None")
    else:
        logger.debug("Credentials loaded from function arguments")
    return credential


def get_and_check_username_password(
    username: Optional[str],
    password: Optional[str],
    credentials_file: Optional[pathlib.Path],
    no_metadata_cache: bool,
) -> Tuple[str, str]:
    username, password = get_username_password(
        username=username, password=password, credentials_file=credentials_file
    )
    copernicus_marine_credentials_are_valid = (
        _are_copernicus_marine_credentials_valid(
            username, password, ignore_cache=no_metadata_cache
        )
    )
    if not copernicus_marine_credentials_are_valid:
        raise InvalidUsernameOrPassword(
            "Learn how to recover your credentials at: "
            "https://help.marine.copernicus.eu/en/articles/"
            "4444552-i-forgot-my-username-or-my-password-what-should-i-do"
        )
    return (username, password)


def get_username_password(
    username: Optional[str],
    password: Optional[str],
    credentials_file: Optional[pathlib.Path],
) -> Tuple[str, str]:
    username = get_credential(
        username,
        "username",
        hide_input=False,
        credentials_file=credentials_file,
    )
    password = get_credential(
        password,
        "password",
        hide_input=True,
        credentials_file=credentials_file,
    )
    return (username, password)


def _get_credential_from_environment_variable_or_prompt(
    credential: Optional[str],
    credential_type: Literal["username", "password"],
    hide_input: bool,
) -> str:
    if not credential:
        credential = _retrieve_credential_from_environment_variable(
            credential_type
        )
        if not credential:
            credential = _retrieve_credential_from_prompt(
                credential_type, hide_input
            )
            if not credential:
                raise CredentialCannotBeNone(credential_type)
    return credential


def credentials_file_builder(
    username: Optional[str],
    password: Optional[str],
    configuration_file_directory: pathlib.Path,
    overwrite_configuration_file: bool,
) -> Optional[pathlib.Path]:
    username = _get_credential_from_environment_variable_or_prompt(
        username, "username", False
    )
    password = _get_credential_from_environment_variable_or_prompt(
        password, "password", True
    )
    copernicus_marine_credentials_are_valid = (
        _are_copernicus_marine_credentials_valid(
            username, password, ignore_cache=False
        )
    )
    if copernicus_marine_credentials_are_valid:
        configuration_file = create_copernicusmarine_configuration_file(
            username=username,
            password=password,
            configuration_file_directory=configuration_file_directory,
            overwrite_configuration_file=overwrite_configuration_file,
        )
        return configuration_file
    return None



import asyncio
import inspect
import warnings
from datetime import datetime
from typing import Awaitable, Callable, Optional, Tuple, TypeVar, Union


# Copyright from: https://github.com/kaelzhang/python-aioretry
class RetryInfo:
    __slots__ = ("fails", "exception", "since")

    fails: int
    exception: Exception
    since: datetime

    def __init__(
        self, fails: int, exception: Exception, since: datetime
    ) -> None:
        self.fails = fails
        self.exception = exception
        self.since = since

    def update(self, exception: Exception) -> "RetryInfo":
        """Create a new RetryInfo and update fails and exception

        Why?
            The object might be collected by user,
            so we need to create a new object every time it fails.
        """

        return RetryInfo(self.fails + 1, exception, self.since)


RetryPolicyStrategy = Tuple[bool, Union[int, float]]

RetryPolicy = Callable[[RetryInfo], RetryPolicyStrategy]
BeforeRetry = Callable[[RetryInfo], Optional[Awaitable[None]]]

ParamRetryPolicy = Union[RetryPolicy, str]
ParamBeforeRetry = Union[BeforeRetry, str]

TargetFunction = Callable[..., Awaitable]
Exceptions = Tuple[Exception, ...]
ExceptionsOrException = Union[Exceptions, Exception]

T = TypeVar("T", RetryPolicy, BeforeRetry)


async def await_coro(coro):
    if inspect.isawaitable(coro):
        return await coro

    return coro


def warn(method_name: str, exception: Exception):
    warnings.warn(
        f"""[aioretry] {method_name} raises an exception:
    {exception}
It is usually a bug that you should fix!""",
        UserWarning,
        stacklevel=2,
    )


async def perform(
    fn: TargetFunction,
    retry_policy: RetryPolicy,
    before_retry: Optional[BeforeRetry],
    *args,
    **kwargs,
):
    info = None

    while True:
        try:
            return await fn(*args, **kwargs)
        except Exception as e:
            if info is None:
                info = RetryInfo(1, e, datetime.now())
            else:
                info = info.update(e)

            try:
                abandon, delay = retry_policy(info)
            except Exception as e2:
                warn("retry_policy", e2)
                raise e2

            if abandon:
                raise e

            if before_retry is not None:
                try:
                    await await_coro(before_retry(info))
                except Exception as e:
                    warn("before_retry", e)
                    raise e

            # `delay` could be 0
            if delay > 0:
                await asyncio.sleep(delay)


def get_method(
    target: Union[T, str],
    args: Tuple,
    name: str,
) -> T:
    if not isinstance(target, str):
        return target

    if len(args) == 0:
        raise RuntimeError(
            f"[aioretry] decorator should be used for instance method"
            f" if {name} as a str '{target}', which should be fixed"
        )

    self = args[0]

    return getattr(self, target)  # type: ignore


def retry(
    retry_policy: ParamRetryPolicy,
    before_retry: Optional[ParamBeforeRetry] = None,
) -> Callable[[TargetFunction], TargetFunction]:
    """Creates a decorator function

    Args:
        retry_policy (RetryPolicy, str): the retry policy
        before_retry (BeforeRetry, str, None): the function to
            be called after each failure of fn
            and before the corresponding retry.

    Returns:
        A wrapped function which accepts the same arguments as
        fn and returns an Awaitable

    Usage::
        @retry(retry_policy)
        async def coro_func():
            ...
    """

    def wrapper(fn: TargetFunction) -> TargetFunction:
        async def wrapped(*args, **kwargs):
            return await perform(
                fn,
                get_method(retry_policy, args, "retry_policy"),
                (
                    get_method(before_retry, args, "before_retry")
                    if before_retry is not None
                    else None
                ),
                *args,
                **kwargs,
            )

        return wrapped

    return wrapper



import pathlib
from typing import Optional

from copernicusmarine.core_functions.deprecated import deprecated_python_option
from copernicusmarine.core_functions.deprecated_options import (
    DEPRECATED_OPTIONS,
)
from copernicusmarine.core_functions.login import login_function
from copernicusmarine.core_functions.utils import DEFAULT_CLIENT_BASE_DIRECTORY


@deprecated_python_option(DEPRECATED_OPTIONS)
def login(
    username: Optional[str] = None,
    password: Optional[str] = None,
    configuration_file_directory: pathlib.Path = DEFAULT_CLIENT_BASE_DIRECTORY,
    overwrite_configuration_file: bool = False,
    skip_if_user_logged_in: bool = False,
) -> bool:
    """
    Create a configuration file with your Copernicus Marine credentials.

    Args:
        username (str, optional): If not set, search for environment variable COPERNICUSMARINE_SERVICE_USERNAME, or else ask for user input.
        password (str, optional): If not set, search for environment variable COPERNICUSMARINE_SERVICE_PASSWORD, or else ask for user input.
        configuration_file_directory (Union[pathlib.Path, str]): Path to the directory where the configuration file is stored.
        overwrite_configuration_file (bool): Flag to skip confirmation before overwriting configuration file.
        skip_if_user_logged_in (bool): Flag to skip the logging process if the user is already logged in.
    """  # noqa
    return login_function(
        username=username,
        password=password,
        configuration_file_directory=configuration_file_directory,
        overwrite_configuration_file=overwrite_configuration_file,
        skip_if_user_logged_in=skip_if_user_logged_in,
    )



import pathlib
from datetime import datetime
from typing import List, Optional, Union

from copernicusmarine.core_functions.deprecated import deprecated_python_option
from copernicusmarine.core_functions.deprecated_options import (
    DEPRECATED_OPTIONS,
)
from copernicusmarine.core_functions.models import (
    DEFAULT_FILE_FORMAT,
    DEFAULT_SUBSET_METHOD,
    FileFormat,
    SubsetMethod,
)
from copernicusmarine.core_functions.subset import subset_function
from copernicusmarine.python_interface.exception_handler import (
    log_exception_and_exit,
)
from copernicusmarine.python_interface.utils import homogenize_datetime


@deprecated_python_option(DEPRECATED_OPTIONS)
@log_exception_and_exit
def subset(
    dataset_url: Optional[str] = None,
    dataset_id: Optional[str] = None,
    dataset_version: Optional[str] = None,
    dataset_part: Optional[str] = None,
    username: Optional[str] = None,
    password: Optional[str] = None,
    variables: Optional[List[str]] = None,
    minimum_longitude: Optional[float] = None,
    maximum_longitude: Optional[float] = None,
    minimum_latitude: Optional[float] = None,
    maximum_latitude: Optional[float] = None,
    minimum_depth: Optional[float] = None,
    maximum_depth: Optional[float] = None,
    vertical_dimension_as_originally_produced: bool = True,
    start_datetime: Optional[Union[datetime, str]] = None,
    end_datetime: Optional[Union[datetime, str]] = None,
    subset_method: SubsetMethod = DEFAULT_SUBSET_METHOD,
    output_filename: Optional[str] = None,
    file_format: FileFormat = DEFAULT_FILE_FORMAT,
    service: Optional[str] = None,
    request_file: Optional[Union[pathlib.Path, str]] = None,
    output_directory: Optional[Union[pathlib.Path, str]] = None,
    credentials_file: Optional[Union[pathlib.Path, str]] = None,
    motu_api_request: Optional[str] = None,
    force_download: bool = False,
    overwrite_output_data: bool = False,
    overwrite_metadata_cache: bool = False,
    no_metadata_cache: bool = False,
    disable_progress_bar: bool = False,
    staging: bool = False,
    netcdf_compression_enabled: bool = False,
    netcdf_compression_level: Optional[int] = None,
    netcdf3_compatible: bool = False,
) -> pathlib.Path:
    """
    Extracts a subset of data from a specified dataset using given parameters.

    Args:
        dataset_url (str, optional): The URL of the dataset to retrieve.
        dataset_id (str, optional): The unique identifier of the dataset.
        dataset_version (str, optional): Force the use of a specific dataset version.
        dataset_part (str, optional): Force the use of a specific dataset part.
        username (str, optional): The username for authentication.
        password (str, optional): The password for authentication.
        output_directory (Union[pathlib.Path, str], optional): The directory where downloaded files will be saved.
        credentials_file (Union[pathlib.Path, str], optional): Path to a file containing authentication credentials.
        force_download (bool, optional): Skip confirmation before download.
        overwrite_output_data (bool, optional): If True, overwrite existing output files.
        request_file (Union[pathlib.Path, str], optional): Path to a file containing request parameters. For more information please refer to the README.
        service (str, optional): Force the use of a specific service.
        overwrite_metadata_cache (bool, optional): If True, overwrite the metadata cache.
        no_metadata_cache (bool, optional): If True, do not use the metadata cache.
        variables (List[str], optional): List of variable names to extract.
        minimum_longitude (float, optional): Minimum longitude value for spatial subset.
        maximum_longitude (float, optional): Maximum longitude value for spatial subset.
        minimum_latitude (float, optional): Minimum latitude value for spatial subset.
        maximum_latitude (float, optional): Maximum latitude value for spatial subset.
        minimum_depth (float, optional): Minimum depth value for vertical subset.
        maximum_depth (float, optional): Maximum depth value for vertical subset.
        vertical_dimension_as_originally_produced (bool, optional): Use original vertical dimension.
        start_datetime (datetime, optional): Start datetime for temporal subset.
        end_datetime (datetime, optional): End datetime for temporal subset.
        subset_method (str, optional): The subset method ('nearest' or 'strict') when requesting the dataset. If strict, you can only request dimension strictly inside the dataset.
        output_filename (str, optional): Output filename for the subsetted data.
        file_format (str, optional): Extension format for the filename.
        motu_api_request (str, optional): MOTU API request string.
        netcdf_compression_enabled (bool, optional): Enable compression level 1 to the NetCDF output file. Use 'netcdf_compression_level' option to customize the compression level.
        netcdf_compression_level (int, optional): Specify a compression level to apply on the NetCDF output file. A value of 0 means no compression, and 9 is the highest level of compression available.
        netcdf3_compatible (bool, optional): Enable downloading the dataset in a netCDF 3 compatible format.
    Returns:
        pathlib.Path: Path to the generated subsetted data file.
    """  # noqa
    request_file = pathlib.Path(request_file) if request_file else None
    output_directory = (
        pathlib.Path(output_directory) if output_directory else None
    )
    credentials_file = (
        pathlib.Path(credentials_file) if credentials_file else None
    )

    start_datetime = homogenize_datetime(start_datetime)
    end_datetime = homogenize_datetime(end_datetime)

    return subset_function(
        dataset_url,
        dataset_id,
        dataset_version,
        dataset_part,
        username,
        password,
        variables,
        minimum_longitude,
        maximum_longitude,
        minimum_latitude,
        maximum_latitude,
        minimum_depth,
        maximum_depth,
        vertical_dimension_as_originally_produced,
        start_datetime,
        end_datetime,
        subset_method,
        output_filename,
        file_format,
        service,
        request_file,
        output_directory,
        credentials_file,
        motu_api_request,
        force_download,
        overwrite_output_data,
        overwrite_metadata_cache,
        no_metadata_cache,
        disable_progress_bar,
        staging=staging,
        netcdf_compression_enabled=netcdf_compression_enabled,
        netcdf_compression_level=netcdf_compression_level,
        netcdf3_compatible=netcdf3_compatible,
    )



from datetime import datetime
from typing import Optional, Union

from copernicusmarine.core_functions.utils import datetime_parser


def homogenize_datetime(
    input_datetime: Optional[Union[datetime, str]]
) -> Optional[datetime]:
    if isinstance(input_datetime, str):
        return datetime_parser(input_datetime)
    return input_datetime



import json
import logging

from copernicusmarine.catalogue_parser.catalogue_parser import (
    CopernicusMarineCatalogue,
    CopernicusMarineDatasetServiceType,
    filter_catalogue_with_strings,
    parse_catalogue,
)
from copernicusmarine.core_functions.utils import (
    create_cache_directory,
    delete_cache_folder,
)
from copernicusmarine.core_functions.versions_verifier import VersionVerifier

logger = logging.getLogger("copernicus_marine_root_logger")


def describe_function(
    include_description: bool,
    include_datasets: bool,
    include_keywords: bool,
    include_versions: bool,
    contains: list[str],
    overwrite_metadata_cache: bool,
    no_metadata_cache: bool,
    disable_progress_bar: bool,
    staging: bool,
) -> str:
    VersionVerifier.check_version_describe(staging)
    if staging:
        logger.warning(
            "Detecting staging flag for describe command. "
            "Data will come from the staging environment."
        )

    if overwrite_metadata_cache:
        delete_cache_folder(quiet=True)

    if not no_metadata_cache:
        create_cache_directory()

    base_catalogue: CopernicusMarineCatalogue = parse_catalogue(
        no_metadata_cache=no_metadata_cache,
        disable_progress_bar=disable_progress_bar,
        staging=staging,
    )
    if not include_versions:
        base_catalogue.filter_only_official_versions_and_parts()

    catalogue_dict = (
        filter_catalogue_with_strings(base_catalogue, contains)
        if contains
        else base_catalogue.__dict__
    )

    def default_filter(obj):
        if isinstance(obj, CopernicusMarineDatasetServiceType):
            return obj.to_json_dict()

        attributes = obj.__dict__
        attributes.pop("__objclass__", None)
        if not include_description:
            attributes.pop("description", None)
        if not include_datasets:
            attributes.pop("datasets", None)
        if not include_keywords:
            attributes.pop("keywords", None)
        return obj.__dict__

    json_dump = json.dumps(
        catalogue_dict, default=default_filter, sort_keys=False, indent=2
    )
    return json_dump



import logging
import pathlib
from typing import Optional

import xarray
import zarr
from dask.delayed import Delayed
from tqdm.dask import TqdmCallback

from copernicusmarine.core_functions.exceptions import (
    NetCDFCompressionNotAvailable,
)

logger = logging.getLogger("copernicus_marine_root_logger")


def get_delayed_download(
    dataset: xarray.Dataset,
    output_path: pathlib.Path,
    netcdf_compression_enabled: bool,
    netcdf_compression_level: Optional[int],
    netcdf3_compatible: bool,
):
    if output_path.suffix == ".zarr":
        if netcdf_compression_enabled:
            raise NetCDFCompressionNotAvailable(
                "--netcdf-compression-enabled option cannot be used when "
                "writing to ZARR"
            )
        delayed = _prepare_download_dataset_as_zarr(dataset, output_path)
    else:
        delayed = _prepare_download_dataset_as_netcdf(
            dataset,
            output_path,
            netcdf_compression_enabled,
            netcdf_compression_level,
            netcdf3_compatible,
        )
    return delayed


def download_delayed_dataset(
    delayed: Delayed, disable_progress_bar: bool
) -> None:
    if disable_progress_bar:
        delayed.compute()
    else:
        with TqdmCallback():
            delayed.compute()


def _prepare_download_dataset_as_netcdf(
    dataset: xarray.Dataset,
    output_path: pathlib.Path,
    netcdf_compression_enabled: bool,
    netcdf_compression_level: Optional[int],
    netcdf3_compatible: bool,
):
    logger.debug("Writing dataset to NetCDF")
    for coord in dataset.coords:
        dataset[coord].encoding["_FillValue"] = None
    if netcdf_compression_enabled:
        complevel = (
            1 if netcdf_compression_level is None else netcdf_compression_level
        )
        logger.info(f"NetCDF compression enabled with level {complevel}")
        comp = dict(
            zlib=True,
            complevel=complevel,
            contiguous=False,
            shuffle=True,
        )
        encoding = {var: comp for var in dataset.data_vars}
    else:
        encoding = None

    xarray_download_format = "NETCDF3_CLASSIC" if netcdf3_compatible else None
    return dataset.to_netcdf(
        output_path,
        mode="w",
        compute=False,
        encoding=encoding,
        format=xarray_download_format,
    )


def _prepare_download_dataset_as_zarr(
    dataset: xarray.Dataset, output_path: pathlib.Path
):
    logger.debug("Writing dataset to Zarr")
    store = zarr.DirectoryStore(output_path)
    return dataset.to_zarr(store=store, mode="w", compute=False)



import ssl
from typing import Any, List, Literal, Optional, Tuple

import aiohttp
import boto3
import botocore
import botocore.config
import certifi
import nest_asyncio
import requests
from requests.adapters import HTTPAdapter, Retry

from copernicusmarine.core_functions.environment_variables import (
    COPERNICUSMARINE_DISABLE_SSL_CONTEXT,
    COPERNICUSMARINE_TRUST_ENV,
    PROXY_HTTP,
    PROXY_HTTPS,
)
from copernicusmarine.core_functions.utils import create_custom_query_function

TRUST_ENV = COPERNICUSMARINE_TRUST_ENV == "True"
PROXIES = {}
if PROXY_HTTP:
    PROXIES["http"] = PROXY_HTTP
if PROXY_HTTPS:
    PROXIES["https"] = PROXY_HTTPS


def _get_ssl_context() -> Optional[ssl.SSLContext]:
    if COPERNICUSMARINE_DISABLE_SSL_CONTEXT is not None:
        return None
    return ssl.create_default_context(cafile=certifi.where())


def get_configured_aiohttp_session() -> aiohttp.ClientSession:
    nest_asyncio.apply()
    connector = aiohttp.TCPConnector(ssl=_get_ssl_context())
    return aiohttp.ClientSession(connector=connector, trust_env=TRUST_ENV)


def get_https_proxy() -> Optional[str]:
    return PROXIES.get("https")


def get_configured_boto3_session(
    endpoint_url: str,
    operation_type: List[Literal["ListObjects", "HeadObject", "GetObject"]],
    username: Optional[str] = None,
    return_ressources: bool = False,
) -> Tuple[Any, Any]:
    config_boto3 = botocore.config.Config(
        s3={"addressing_style": "virtual"},
        signature_version=botocore.UNSIGNED,
        retries={"max_attempts": 10, "mode": "standard"},
    )
    s3_session = boto3.Session()
    s3_client = s3_session.client(
        "s3",
        config=config_boto3,
        endpoint_url=endpoint_url,
    )
    for operation in operation_type:
        # Register the botocore event handler for adding custom query params
        # to S3 HEAD and GET requests
        s3_client.meta.events.register(
            f"before-call.s3.{operation}",
            create_custom_query_function(username),
        )
    if not return_ressources:
        return s3_client, None
    s3_resource = boto3.resource(
        "s3",
        config=config_boto3,
        endpoint_url=endpoint_url,
    )
    return s3_client, s3_resource


def get_configured_requests_session() -> requests.Session:
    session = requests.Session()
    session.trust_env = TRUST_ENV
    session.verify = certifi.where()
    session.proxies = PROXIES
    session.mount(
        "https://",
        HTTPAdapter(
            max_retries=Retry(
                total=5,
                backoff_factor=1,
                status_forcelist=[500, 502, 503, 504],
            )
        ),
    )
    return session



import logging
from dataclasses import dataclass
from enum import Enum
from itertools import chain
from typing import List, Literal, Optional, Tuple, Union

from copernicusmarine.catalogue_parser.catalogue_parser import (
    CopernicusMarineCatalogue,
    CopernicusMarineDatasetServiceType,
    CopernicusMarineDatasetVersion,
    CopernicusMarineProductDataset,
    CopernicusMarineService,
    CopernicusMarineServiceFormat,
    CopernicusMarineVersionPart,
)
from copernicusmarine.catalogue_parser.request_structure import (
    DatasetTimeAndGeographicalSubset,
)
from copernicusmarine.core_functions import custom_open_zarr
from copernicusmarine.core_functions.utils import (
    FormatNotSupported,
    datetime_parser,
    next_or_raise_exception,
)
from copernicusmarine.download_functions.subset_xarray import (
    get_size_of_coordinate_subset,
)

logger = logging.getLogger("copernicus_marine_root_logger")


class _Command(Enum):
    GET = "get"
    SUBSET = "subset"
    LOAD = "load"


@dataclass(frozen=True)
class Command:
    command_name: _Command
    service_types_by_priority: List[CopernicusMarineDatasetServiceType]

    def service_names(self) -> List[str]:
        return list(
            map(
                lambda service_type: service_type.service_name.value,
                self.service_types_by_priority,
            )
        )

    def service_short_names(self) -> List[str]:
        return list(
            map(
                lambda service_type: service_type.short_name.value,
                self.service_types_by_priority,
            )
        )

    def service_aliases(self) -> List[str]:
        return list(
            chain(
                *map(
                    lambda service_type: service_type.aliases(),
                    self.service_types_by_priority,
                )
            )
        )


class CommandType(Command, Enum):
    SUBSET = (
        _Command.SUBSET,
        [
            CopernicusMarineDatasetServiceType.GEOSERIES,
            CopernicusMarineDatasetServiceType.TIMESERIES,
            CopernicusMarineDatasetServiceType.OMI_ARCO,
            CopernicusMarineDatasetServiceType.STATIC_ARCO,
        ],
    )
    GET = (
        _Command.GET,
        [
            CopernicusMarineDatasetServiceType.FILES,
        ],
    )
    LOAD = (
        _Command.LOAD,
        [
            CopernicusMarineDatasetServiceType.GEOSERIES,
            CopernicusMarineDatasetServiceType.TIMESERIES,
            CopernicusMarineDatasetServiceType.OMI_ARCO,
            CopernicusMarineDatasetServiceType.STATIC_ARCO,
        ],
    )


def assert_service_type_for_command(
    service_type: CopernicusMarineDatasetServiceType, command_type: CommandType
) -> CopernicusMarineDatasetServiceType:
    return next_or_raise_exception(
        (
            service_type
            for service_type in command_type.service_types_by_priority
        ),
        _service_type_does_not_exist_for_command(service_type, command_type),
    )


class ServiceDoesNotExistForCommand(Exception):
    def __init__(self, service_name, command_name, available_services):
        super().__init__()
        self.__setattr__(
            "custom_exception_message",
            f"Service {service_name} "
            f"does not exist for command {command_name}. "
            f"Possible service{'s' if len(available_services) > 1 else ''}: "
            f"{available_services}",
        )


def _service_type_does_not_exist_for_command(
    service_type: CopernicusMarineDatasetServiceType, command_type: CommandType
) -> ServiceDoesNotExistForCommand:
    return _service_does_not_exist_for_command(
        service_type.service_name.value, command_type
    )


def _service_does_not_exist_for_command(
    service_name: str, command_type: CommandType
) -> ServiceDoesNotExistForCommand:
    return ServiceDoesNotExistForCommand(
        service_name,
        command_type.command_name.value,
        command_type.service_aliases(),
    )


def _select_forced_service(
    dataset_version_part: CopernicusMarineVersionPart,
    force_service_type: CopernicusMarineDatasetServiceType,
    command_type: CommandType,
) -> CopernicusMarineService:
    return next_or_raise_exception(
        (
            service
            for service in dataset_version_part.services
            if service.service_type == force_service_type
        ),
        _service_not_available_error(dataset_version_part, command_type),
    )


def _get_best_arco_service_type(
    dataset_subset: DatasetTimeAndGeographicalSubset,
    dataset_url: str,
    username: Optional[str],
) -> Literal[
    CopernicusMarineDatasetServiceType.TIMESERIES,
    CopernicusMarineDatasetServiceType.GEOSERIES,
]:
    dataset = custom_open_zarr.open_zarr(
        dataset_url, copernicus_marine_username=username
    )

    latitude_size = get_size_of_coordinate_subset(
        dataset,
        "latitude",
        dataset_subset.minimum_latitude,
        dataset_subset.maximum_latitude,
    )
    longitude_size = get_size_of_coordinate_subset(
        dataset,
        "longitude",
        dataset_subset.minimum_longitude,
        dataset_subset.maximum_longitude,
    )
    time_size = get_size_of_coordinate_subset(
        dataset,
        "time",
        dataset_subset.start_datetime,
        dataset_subset.end_datetime,
    )
    dataset_coordinates = dataset.coords

    geographical_dimensions = (
        dataset_coordinates["latitude"].size
        * dataset_coordinates["longitude"].size
    )
    subset_geographical_dimensions = latitude_size * longitude_size
    temporal_dimensions = dataset_coordinates["time"].size
    subset_temporal_dimensions = time_size

    geographical_coverage = (
        subset_geographical_dimensions / geographical_dimensions
    )
    temporal_coverage = subset_temporal_dimensions / temporal_dimensions

    if geographical_coverage >= temporal_coverage:
        return CopernicusMarineDatasetServiceType.GEOSERIES
    return CopernicusMarineDatasetServiceType.TIMESERIES


def _get_first_available_service_type(
    command_type: CommandType,
    dataset_available_service_types: list[CopernicusMarineDatasetServiceType],
) -> CopernicusMarineDatasetServiceType:
    available_service_types = command_type.service_types_by_priority
    return next_or_raise_exception(
        (
            service_type
            for service_type in available_service_types
            if service_type in dataset_available_service_types
        ),
        _no_service_available_for_command(command_type),
    )


def _select_service_by_priority(
    dataset_version_part: CopernicusMarineVersionPart,
    command_type: CommandType,
    dataset_subset: Optional[DatasetTimeAndGeographicalSubset],
    username: Optional[str],
) -> CopernicusMarineService:
    dataset_available_service_types = [
        service.service_type for service in dataset_version_part.services
    ]
    first_available_service_type = _get_first_available_service_type(
        command_type=command_type,
        dataset_available_service_types=dataset_available_service_types,
    )
    first_available_service = dataset_version_part.get_service_by_service_type(
        service_type=first_available_service_type
    )
    if (
        CopernicusMarineDatasetServiceType.GEOSERIES
        in dataset_available_service_types
        and CopernicusMarineDatasetServiceType.TIMESERIES
        in dataset_available_service_types
        and command_type in [CommandType.SUBSET, CommandType.LOAD]
        and dataset_subset is not None
    ):
        if (
            first_available_service.service_format
            == CopernicusMarineServiceFormat.SQLITE
        ):
            raise FormatNotSupported(
                first_available_service.service_format.value
            )
        best_arco_service_type: CopernicusMarineDatasetServiceType = (
            _get_best_arco_service_type(
                dataset_subset, first_available_service.uri, username
            )
        )
        return dataset_version_part.get_service_by_service_type(
            best_arco_service_type
        )
    return first_available_service


def parse_dataset_id_and_service_and_suffix_path_from_url(
    catalogue: CopernicusMarineCatalogue,
    dataset_url: Optional[str],
) -> Tuple[str, CopernicusMarineDatasetServiceType, str,]:
    if dataset_url is None:
        syntax_error = SyntaxError(
            "Must specify at least one of "
            "'dataset_url' or 'dataset_id' options"
        )
        raise syntax_error
    return next_or_raise_exception(
        (
            (
                dataset.dataset_id,
                service.service_type,
                dataset_url.split(service.uri)[1],
            )
            for product in catalogue.products
            for dataset in product.datasets
            for dataset_version in dataset.versions
            for dataset_part in dataset_version.parts
            for service in dataset_part.services
            if dataset_url.startswith(service.uri)
        ),
        KeyError(
            f"The requested dataset URL '{dataset_url}' "
            "was not found in the catalogue, "
            "you can use 'copernicusmarine describe --include-datasets "
            "--contains <search_token>' to find datasets"
        ),
    )


@dataclass
class RetrievalService:
    dataset_id: str
    service_type: CopernicusMarineDatasetServiceType
    service_format: Optional[CopernicusMarineServiceFormat]
    uri: str
    dataset_valid_start_date: Optional[Union[str, int]]


def get_retrieval_service(
    catalogue: CopernicusMarineCatalogue,
    dataset_id: Optional[str],
    dataset_url: Optional[str],
    force_dataset_version_label: Optional[str],
    force_dataset_part_label: Optional[str],
    force_service_type_string: Optional[str],
    command_type: CommandType,
    index_parts: bool = False,
    dataset_subset: Optional[DatasetTimeAndGeographicalSubset] = None,
    dataset_sync: bool = False,
    username: Optional[str] = None,
) -> RetrievalService:
    force_service_type: Optional[CopernicusMarineDatasetServiceType] = (
        _service_type_from_string(force_service_type_string, command_type)
        if force_service_type_string
        else None
    )
    if dataset_id is None:
        (
            dataset_id,
            service_type,
            suffix_path,
        ) = parse_dataset_id_and_service_and_suffix_path_from_url(
            catalogue, dataset_url
        )
        force_service_type = (
            service_type if not force_service_type else force_service_type
        )
    else:
        if dataset_url is not None:
            syntax_error = SyntaxError(
                "Must specify only one of 'dataset_url' or 'dataset_id' options"
            )
            raise syntax_error
        suffix_path = ""

    return _get_retrieval_service_from_dataset_id(
        catalogue=catalogue,
        dataset_id=dataset_id,
        suffix_path=suffix_path,
        force_dataset_version_label=force_dataset_version_label,
        force_dataset_part_label=force_dataset_part_label,
        force_service_type=force_service_type,
        command_type=command_type,
        index_parts=index_parts,
        dataset_subset=dataset_subset,
        dataset_sync=dataset_sync,
        username=username,
    )


def _get_retrieval_service_from_dataset_id(
    catalogue: CopernicusMarineCatalogue,
    dataset_id: str,
    suffix_path: str,
    force_dataset_version_label: Optional[str],
    force_dataset_part_label: Optional[str],
    force_service_type: Optional[CopernicusMarineDatasetServiceType],
    command_type: CommandType,
    index_parts: bool,
    dataset_subset: Optional[DatasetTimeAndGeographicalSubset],
    dataset_sync: bool,
    username: Optional[str],
) -> RetrievalService:
    dataset: CopernicusMarineProductDataset = next_or_raise_exception(
        (
            dataset
            for product in catalogue.products
            for dataset in product.datasets
            if dataset_id == dataset.dataset_id
        ),
        KeyError(
            f"The requested dataset '{dataset_id}' was not found in the catalogue,"
            " you can use 'copernicusmarine describe --include-datasets "
            "--contains <search_token>' to find datasets"
        ),
    )
    return _get_retrieval_service_from_dataset(
        dataset=dataset,
        suffix_path=suffix_path,
        force_dataset_version_label=force_dataset_version_label,
        force_dataset_part_label=force_dataset_part_label,
        force_service_type=force_service_type,
        command_type=command_type,
        index_parts=index_parts,
        dataset_subset=dataset_subset,
        dataset_sync=dataset_sync,
        username=username,
    )


def _get_retrieval_service_from_dataset(
    dataset: CopernicusMarineProductDataset,
    suffix_path: str,
    force_dataset_version_label: Optional[str],
    force_dataset_part_label: Optional[str],
    force_service_type: Optional[CopernicusMarineDatasetServiceType],
    command_type: CommandType,
    index_parts: bool,
    dataset_subset: Optional[DatasetTimeAndGeographicalSubset],
    dataset_sync: bool,
    username: Optional[str],
) -> RetrievalService:
    if force_dataset_version_label:
        logger.info(
            "You forced selection of dataset version "
            + f'"{force_dataset_version_label}"'
        )
    dataset_version = dataset.get_version(force_dataset_version_label)
    if not force_dataset_version_label:
        logger.info(
            "Dataset version was not specified, the latest "
            f'one was selected: "{dataset_version.label}"'
        )
    return _get_retrieval_service_from_dataset_version(
        dataset_id=dataset.dataset_id,
        dataset_version=dataset_version,
        force_dataset_part_label=force_dataset_part_label,
        suffix_path=suffix_path,
        force_service_type=force_service_type,
        command_type=command_type,
        index_parts=index_parts,
        dataset_subset=dataset_subset,
        dataset_sync=dataset_sync,
        username=username,
    )


def _get_retrieval_service_from_dataset_version(
    dataset_id: str,
    dataset_version: CopernicusMarineDatasetVersion,
    force_dataset_part_label: Optional[str],
    suffix_path: str,
    force_service_type: Optional[CopernicusMarineDatasetServiceType],
    command_type: CommandType,
    index_parts: bool,
    dataset_subset: Optional[DatasetTimeAndGeographicalSubset],
    dataset_sync: bool,
    username: Optional[str],
) -> RetrievalService:
    if len(dataset_version.parts) > 1 and dataset_sync:
        raise Exception(
            "Sync is not supported for datasets with multiple parts."
        )
    if (
        force_service_type == CopernicusMarineDatasetServiceType.FILES
        and not force_dataset_part_label
        and not index_parts
        and len(dataset_version.parts) > 1
    ):
        raise Exception(
            "When dataset has multiple parts and using 'files' service"
            ", please indicate the part you want to download "
            "with the dataset-part option"
        )
    if force_dataset_part_label:
        logger.info(
            f"You forced selection of dataset part "
            f'"{force_dataset_part_label}"'
        )
    dataset_part = dataset_version.get_part(force_dataset_part_label)
    if not force_dataset_part_label:
        logger.info(
            "Dataset part was not specified, the first "
            f'one was selected: "{dataset_part.name}"'
        )
    if dataset_part.retired_date:
        _warning_dataset_will_be_deprecated(
            dataset_id, dataset_version, dataset_part
        )
    if dataset_part.released_date and datetime_parser(
        dataset_part.released_date
    ) > datetime_parser("now"):
        _warning_dataset_not_yet_released(
            dataset_id, dataset_version, dataset_part
        )

    if force_service_type:
        logger.info(
            f"You forced selection of service: "
            f"{force_service_type.service_name.value}"
        )
        service = _select_forced_service(
            dataset_version_part=dataset_part,
            force_service_type=force_service_type,
            command_type=command_type,
        )
        if service.service_format == CopernicusMarineServiceFormat.SQLITE:
            raise FormatNotSupported(service.service_format.value)
    else:
        service = _select_service_by_priority(
            dataset_version_part=dataset_part,
            command_type=command_type,
            dataset_subset=dataset_subset,
            username=username,
        )
        logger.info(
            "Service was not specified, the default one was "
            f'selected: "{service.service_type.service_name.value}"'
        )
    dataset_start_date = _get_dataset_start_date_from_service(service)
    return RetrievalService(
        dataset_id=dataset_id,
        service_type=service.service_type,
        uri=service.uri + suffix_path,
        dataset_valid_start_date=dataset_start_date,
        service_format=service.service_format,
    )


def _get_dataset_start_date_from_service(
    service: CopernicusMarineService,
) -> Optional[Union[str, int]]:
    for variable in service.variables:
        for coordinate in variable.coordinates:
            if (
                coordinate.coordinates_id == "time"
                and coordinate.minimum_value
            ):
                if isinstance(coordinate.minimum_value, str):
                    return coordinate.minimum_value.replace("Z", "")
                return int(coordinate.minimum_value)
    return None


class ServiceNotAvailable(Exception):
    ...


def _warning_dataset_will_be_deprecated(
    dataset_id: str,
    dataset_version: CopernicusMarineDatasetVersion,
    dataset_part: CopernicusMarineVersionPart,
):
    logger.warning(
        f"""The dataset {dataset_id}"""
        f"""{f", version '{dataset_version.label}'"
             if dataset_version.label != 'default' else ''}"""
        f"""{(f"and part '{dataset_part.name}'"
              if dataset_part.name != 'default' else '')}"""
        f"""{"," if dataset_version.label != 'default' else ""}"""
        f""" will be retired on the {dataset_part.retired_date}."""
        """ After this date, it will no longer be available on the toolbox."""
    )


def _warning_dataset_not_yet_released(
    dataset_id: str,
    dataset_version: CopernicusMarineDatasetVersion,
    dataset_part: CopernicusMarineVersionPart,
):
    logger.warning(
        f"""The dataset {dataset_id}"""
        f"""{f", version '{dataset_version.label}'"
             if dataset_version.label != 'default' else ''}"""
        f"""{(f"and part '{dataset_part.name}'"
              if dataset_part.name != 'default' else '')}"""
        f"""{"," if dataset_version.label != 'default' else ""}"""
        f""" is not yet released officially."""
        f""" It will be available by default  on the toolbox on the"""
        f""" {dataset_part.released_date}."""
    )


def _service_not_available_error(
    dataset_version_part: CopernicusMarineVersionPart,
    command_type: CommandType,
) -> ServiceNotAvailable:
    dataset_available_service_types = [
        service.service_type.short_name.value
        for service in dataset_version_part.services
        if service.service_type in command_type.service_types_by_priority
    ]
    return ServiceNotAvailable(
        f"Available services for dataset: "
        f"{dataset_available_service_types}"
    )


class NoServiceAvailable(Exception):
    ...


def _no_service_available_for_command(
    command_type: CommandType,
) -> NoServiceAvailable:
    return NoServiceAvailable(
        f"No service available for dataset "
        f"with command {command_type.command_name.value}"
    )


def _service_type_from_string(
    string: str, command_type: CommandType
) -> CopernicusMarineDatasetServiceType:
    return next_or_raise_exception(
        (
            service_type
            for service_type in command_type.service_types_by_priority
            if string in service_type.aliases()
        ),
        _service_does_not_exist_for_command(string, command_type),
    )



import logging

import semver

from copernicusmarine import __version__ as client_version
from copernicusmarine.core_functions.sessions import (
    get_configured_requests_session,
)
from copernicusmarine.core_functions.utils import (
    construct_query_params_for_marine_data_store_monitoring,
)

logger = logging.getLogger("copernicus_marine_root_logger")


class VersionVerifier:
    function_marine_data_store_service_mapping: dict[str, list[str]] = {
        "describe": ["mds", "mds/serverlessArco/meta"],
        "get": ["mds", "mds/serverlessNative", "mds/serverlessArco/meta"],
        "subset": ["mds", "mds/serverlessArco", "mds/serverlessArco/meta"],
    }

    @staticmethod
    def check_version_describe(staging: bool):
        VersionVerifier._check_version("describe", staging)

    @staticmethod
    def check_version_get(staging: bool):
        VersionVerifier._check_version("get", staging)

    @staticmethod
    def check_version_subset(staging: bool):
        VersionVerifier._check_version("subset", staging)

    @staticmethod
    def _check_version(function_name: str, staging: bool):
        marine_data_store_versions = (
            VersionVerifier._get_client_required_versions(staging)
        )
        for (
            service
        ) in VersionVerifier.function_marine_data_store_service_mapping[
            function_name
        ]:
            required_version = marine_data_store_versions[service]
            if not semver.Version.parse(client_version).match(
                required_version
            ):
                logger.debug(
                    f"Client version {client_version} is not compatible with "
                    f"{service}. Service needs version {required_version}."
                )
                logger.error(
                    f"Client version {client_version} is not compatible with current "
                    "backend service. Please update to the latest client version."
                )

    @staticmethod
    def _get_client_required_versions(
        staging: bool,
    ) -> dict[str, str]:
        url_mds_versions = (
            "https://s3.waw3-1.cloudferro.com/mdl-metadata-dta/mdsVersions.json"
            if staging
            else "https://s3.waw3-1.cloudferro.com/mdl-metadata/mdsVersions.json"
        )
        logger.debug(f"Getting required versions from {url_mds_versions}")
        session = get_configured_requests_session()
        mds_versions: dict[str, str] = session.get(
            url_mds_versions,
            params=construct_query_params_for_marine_data_store_monitoring(),
            proxies=session.proxies,
        ).json()["clientVersions"]
        return mds_versions



from collections.abc import Iterator, Mapping
from typing import Dict, List, Optional


class DeprecatedOption:
    def __init__(
        self,
        old_name,
        new_name,
        replace=True,
        deprecated_for_v2=False,
        deleted_for_v2=True,
        only_for_v2=False,
        targeted_functions: Optional[list[str]] = None,
    ) -> None:
        self.old_name = old_name
        self.new_name = new_name
        self.replace = replace
        self.deprecated_for_v2 = deprecated_for_v2
        self.deleted_for_v2 = deleted_for_v2
        self.only_for_v2 = only_for_v2
        if not targeted_functions:
            self.targeted_functions = [
                "describe",
                "get",
                "subset",
                "login",
                "open_dataset",
                "read_dataframe",
            ]
        else:
            self.targeted_functions = targeted_functions


class DeprecatedOptionMapping(Mapping):
    def __init__(self, deprecated_options: List[DeprecatedOption]) -> None:
        self.deprecated_options_by_old_names: Dict[str, DeprecatedOption] = {}
        for value in deprecated_options:
            if value not in self.deprecated_options_by_old_names:
                self.deprecated_options_by_old_names[value.old_name] = value

    def __getitem__(self, __key: str) -> DeprecatedOption:
        return self.deprecated_options_by_old_names[__key]

    def __iter__(self) -> Iterator:
        return self.deprecated_options_by_old_names.__iter__()

    def __len__(self) -> int:
        return self.deprecated_options_by_old_names.__len__()

    @property
    def dict_old_names_to_new_names(self):
        result_dict = {}
        for (
            old_name,
            deprecated_option,
        ) in self.deprecated_options_by_old_names.items():
            if deprecated_option.replace:
                result_dict[old_name] = deprecated_option.new_name
        return result_dict


DEPRECATED_OPTIONS: DeprecatedOptionMapping = DeprecatedOptionMapping(
    [
        DeprecatedOption(
            old_name="minimal_longitude", new_name="minimum_longitude"
        ),
        DeprecatedOption(
            old_name="maximal_longitude", new_name="maximum_longitude"
        ),
        DeprecatedOption(
            old_name="minimal_latitude", new_name="minimum_latitude"
        ),
        DeprecatedOption(
            old_name="maximal_latitude", new_name="maximum_latitude"
        ),
        DeprecatedOption(old_name="minimal_depth", new_name="minimum_depth"),
        DeprecatedOption(old_name="maximal_depth", new_name="maximum_depth"),
        DeprecatedOption(
            old_name="force_dataset_version", new_name="dataset_version"
        ),
        DeprecatedOption(
            old_name="force_dataset_part", new_name="dataset_part"
        ),
        DeprecatedOption(old_name="force_service", new_name="service"),
        DeprecatedOption(
            old_name="service",
            new_name="service",
            replace=False,
            deleted_for_v2=True,
            only_for_v2=True,
            targeted_functions=["get"],
        ),
        DeprecatedOption(
            old_name="download_file_list",
            new_name="create_file_list",
            replace=False,
        ),
        DeprecatedOption(
            old_name="include_all_versions",
            new_name="include_versions",
        ),
        DeprecatedOption(
            old_name="skip_if_user_logged_in",
            new_name="skip_if_user_logged_in",
            only_for_v2=True,
        ),
        DeprecatedOption(
            old_name="overwrite_metadata_cache",
            new_name="overwrite_metadata_cache",
            only_for_v2=True,
        ),
        DeprecatedOption(
            old_name="no_metadata_cache",
            new_name="no_metadata_cache",
            only_for_v2=True,
        ),
        DeprecatedOption(
            old_name="dataset_url",
            new_name="dataset_url",
            only_for_v2=True,
        ),
        DeprecatedOption(
            old_name="subset_method",
            new_name="coordinates_selection_method",
            replace=False,
            only_for_v2=True,
        ),
        DeprecatedOption(
            old_name="netcdf_compression_enabled",
            new_name="netcdf_compression_enabled",
            only_for_v2=True,
        ),
        DeprecatedOption(
            old_name="disable_progress_bar",
            new_name="disable_progress_bar",
            targeted_functions=["open_dataset", "read_dataframe"],
            only_for_v2=True,
        ),
        DeprecatedOption(
            old_name="include_datasets",
            new_name="include_datasets",
            deprecated_for_v2=True,
            deleted_for_v2=False,
            only_for_v2=True,
        ),
        DeprecatedOption(
            old_name="include_description",
            new_name="include_description",
            deprecated_for_v2=True,
            deleted_for_v2=False,
            only_for_v2=True,
        ),
        DeprecatedOption(
            old_name="include_keywords",
            new_name="include_keywords",
            deprecated_for_v2=True,
            deleted_for_v2=False,
            only_for_v2=True,
        ),
        DeprecatedOption(
            old_name="include_all",
            new_name="include_all",
            deprecated_for_v2=True,
            deleted_for_v2=False,
            only_for_v2=True,
        ),
    ]
)
